from django import forms
from django.utils.translation import gettext, pgettext_lazy

from ....admin.forms import ColorField, YesNoSwitch
from ...models import SocialAuthProvider
from ..ordering import get_next_free_order


class ProviderForm(forms.ModelForm):
    button_text = forms.CharField(
        label=pgettext_lazy("admin social auth provider form", "Button text"),
        required=False,
    )
    button_color = ColorField(
        label=pgettext_lazy("admin social auth provider form", "Button color"),
        required=False,
    )
    is_active = YesNoSwitch(
        label=pgettext_lazy("admin social auth provider form", "Enable this provider")
    )

    class Meta:
        model = SocialAuthProvider
        fields = ["button_text", "button_color", "is_active"]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request")
        super().__init__(*args, **kwargs)

    def save(self):
        settings = {}
        for setting, value in self.cleaned_data.items():
            if setting not in ["button_text", "button_color", "is_active"]:
                settings[setting] = value
        self.instance.settings = settings

        if "is_active" in self.changed_data and self.cleaned_data.get("is_active"):
            self.instance.order = get_next_free_order()

        self.instance.save()


class OAuthProviderForm(ProviderForm):
    associate_by_email = YesNoSwitch(
        label=pgettext_lazy(
            "admin social auth provider form", "Associate existing users by email"
        ),
        help_text=pgettext_lazy(
            "admin social auth provider form",
            "This setting controls handling of e-mail collisions that occur when e-mail address returned from social site is already stored in the forum's database. Enabling this option will result in the user being signed in to pre-existing account. Otherwise they will be asked to specify a different e-mail to complete the sign in using social site.",
        ),
    )

    def clean(self):
        cleaned_data = super().clean()

        if cleaned_data.get("is_active"):
            if not cleaned_data.get("key"):
                self.add_error("key", gettext("This field is required."))
            if not cleaned_data.get("secret"):
                self.add_error("secret", gettext("This field is required."))

        return cleaned_data



from ...users.models import Ban
from ..bans import get_fake_email_ban, get_fake_ip_ban, get_fake_username_ban


def test_fake_username_ban_can_be_created(db, fake):
    assert get_fake_username_ban(fake)
    Ban.objects.get(check_type=Ban.USERNAME)


def test_fake_email_ban_can_be_created(db, fake):
    assert get_fake_email_ban(fake)
    Ban.objects.get(check_type=Ban.EMAIL)


def test_fake_ip_ban_can_be_created(db, fake):
    assert get_fake_ip_ban(fake)
    Ban.objects.get(check_type=Ban.IP)



from django.contrib.auth import authenticate, get_user_model, login
from django.core.exceptions import PermissionDenied, ValidationError
from django.db import IntegrityError
from django.utils.translation import pgettext
from django.views.decorators.csrf import csrf_protect
from rest_framework import status
from rest_framework.response import Response

from ... import captcha
from ....legal.models import Agreement
from ...forms.register import RegisterForm
from ...registration import (
    get_registration_result_json,
    save_user_agreements,
    send_welcome_email,
)
from ...setupnewuser import setup_new_user

User = get_user_model()


@csrf_protect
def create_endpoint(request):
    if request.settings.account_activation == "closed":
        raise PermissionDenied(
            pgettext(
                "user register api", "New users registrations are currently closed."
            )
        )

    request_data = request.data
    if not isinstance(request_data, dict):
        request_data = {}

    form = RegisterForm(
        request_data, request=request, agreements=Agreement.objects.get_agreements()
    )

    try:
        if form.is_valid():
            captcha.test_request(request)
    except ValidationError as e:
        form.add_error("captcha", e)

    if not form.is_valid():
        return Response(form.errors, status=status.HTTP_400_BAD_REQUEST)

    activation_kwargs = {}
    if request.settings.account_activation == "user":
        activation_kwargs = {"requires_activation": User.ACTIVATION_USER}
    elif request.settings.account_activation == "admin":
        activation_kwargs = {"requires_activation": User.ACTIVATION_ADMIN}

    try:
        new_user = User.objects.create_user(
            form.cleaned_data["username"],
            form.cleaned_data["email"],
            form.cleaned_data["password"],
            joined_from_ip=request.user_ip,
            **activation_kwargs
        )
    except IntegrityError:
        return Response(
            {
                "__all__": pgettext(
                    "user register api", "Please try resubmitting the form."
                )
            },
            status=status.HTTP_400_BAD_REQUEST,
        )

    setup_new_user(request.settings, new_user)
    save_user_agreements(new_user, form)
    send_welcome_email(request, new_user)

    if new_user.requires_activation == User.ACTIVATION_NONE:
        authenticated_user = authenticate(
            username=new_user.email, password=form.cleaned_data["password"]
        )
        login(request, authenticated_user)

    return Response(get_registration_result_json(new_user))



import json

from django.urls import reverse

from .. import test
from ...categories.models import Category
from ...conf.test import override_dynamic_settings
from ..models import Thread
from ..test import patch_category_acl, patch_other_category_acl
from .test_threads_api import ThreadsApiTestCase


class ThreadsBulkPatchApiTestCase(ThreadsApiTestCase):
    def setUp(self):
        super().setUp()

        self.threads = list(
            reversed(
                [
                    test.post_thread(category=self.category),
                    test.post_thread(category=self.category),
                    test.post_thread(category=self.category),
                ]
            )
        )

        self.ids = list(reversed([t.id for t in self.threads]))

        self.api_link = reverse("misago:api:thread-list")

    def patch(self, api_link, ops):
        return self.client.patch(
            api_link, json.dumps(ops), content_type="application/json"
        )


class BulkPatchSerializerTests(ThreadsBulkPatchApiTestCase):
    def test_invalid_input_type(self):
        """api rejects invalid input type"""
        response = self.patch(self.api_link, [1, 2, 3])

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "non_field_errors": [
                    "Invalid data. Expected a dictionary, but got list."
                ]
            },
        )

    def test_missing_input_keys(self):
        """api rejects input with missing keys"""
        response = self.patch(self.api_link, {})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"ids": ["This field is required."], "ops": ["This field is required."]},
        )

    def test_empty_input_keys(self):
        """api rejects input with empty keys"""
        response = self.patch(self.api_link, {"ids": [], "ops": []})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "ids": ["Ensure this field has at least 1 elements."],
                "ops": ["Ensure this field has at least 1 elements."],
            },
        )

    def test_invalid_input_keys(self):
        """api rejects input with invalid keys"""
        response = self.patch(self.api_link, {"ids": ["a"], "ops": [1]})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "ids": {"0": ["A valid integer is required."]},
                "ops": {"0": ['Expected a dictionary of items but got type "int".']},
            },
        )

    def test_too_small_id(self):
        """api rejects input with implausiple id"""
        response = self.patch(self.api_link, {"ids": [0], "ops": [{}]})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"ids": {"0": ["Ensure this value is greater than or equal to 1."]}},
        )

    @override_dynamic_settings(threads_per_page=5)
    def test_too_large_input(self):
        """api rejects too large input"""
        response = self.patch(
            self.api_link,
            {"ids": [i + 1 for i in range(6)], "ops": [{} for i in range(200)]},
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "ids": ["No more than 5 threads can be updated at a single time."],
                "ops": ["Ensure this field has no more than 10 elements."],
            },
        )

    def test_threads_not_found(self):
        """api fails to find threads"""
        threads = [
            test.post_thread(category=self.category, is_hidden=True),
            test.post_thread(category=self.category, is_unapproved=True),
        ]

        response = self.patch(
            self.api_link, {"ids": [t.id for t in threads], "ops": [{}]}
        )

        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "One or more threads to update could not be found."},
        )

    def test_ops_invalid(self):
        """api validates descriptions"""
        response = self.patch(self.api_link, {"ids": self.ids[:1], "ops": [{}]})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), [{"id": self.ids[0], "detail": ["undefined op"]}]
        )

    def test_anonymous_user(self):
        """anonymous users can't use bulk actions"""
        self.logout_user()

        response = self.patch(self.api_link, {"ids": self.ids[:1], "ops": [{}]})
        self.assertEqual(response.status_code, 403)


class ThreadAddAclApiTests(ThreadsBulkPatchApiTestCase):
    def test_add_acl_true(self):
        """api adds current threads acl to response"""
        response = self.patch(
            self.api_link,
            {"ids": self.ids, "ops": [{"op": "add", "path": "acl", "value": True}]},
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        for i, thread in enumerate(self.threads):
            self.assertEqual(response_json[i]["id"], thread.id)
            self.assertTrue(response_json[i]["acl"])


class BulkThreadChangeTitleApiTests(ThreadsBulkPatchApiTestCase):
    @patch_category_acl({"can_edit_threads": 2})
    def test_change_thread_title(self):
        """api changes thread title and resyncs the category"""
        response = self.patch(
            self.api_link,
            {
                "ids": self.ids,
                "ops": [
                    {"op": "replace", "path": "title", "value": "Changed the title!"}
                ],
            },
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        for i, thread in enumerate(self.threads):
            self.assertEqual(response_json[i]["id"], thread.id)
            self.assertEqual(response_json[i]["title"], "Changed the title!")

        for thread in Thread.objects.filter(id__in=self.ids):
            self.assertEqual(thread.title, "Changed the title!")

        category = Category.objects.get(pk=self.category.id)
        self.assertEqual(category.last_thread_title, "Changed the title!")

    @patch_category_acl({"can_edit_threads": 0})
    def test_change_thread_title_no_permission(self):
        """api validates permission to change title, returns errors"""
        response = self.patch(
            self.api_link,
            {
                "ids": self.ids,
                "ops": [
                    {"op": "replace", "path": "title", "value": "Changed the title!"}
                ],
            },
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        for i, thread in enumerate(self.threads):
            self.assertEqual(response_json[i]["id"], thread.id)
            self.assertEqual(
                response_json[i]["detail"], ["You can't edit threads in this category."]
            )


class BulkThreadMoveApiTests(ThreadsBulkPatchApiTestCase):
    def setUp(self):
        super().setUp()

        Category(name="Other Category", slug="other-category").insert_at(
            self.category, position="last-child", save=True
        )
        self.other_category = Category.objects.get(slug="other-category")

    @patch_category_acl({"can_move_threads": True})
    @patch_other_category_acl({"can_start_threads": 2})
    def test_move_thread(self):
        """api moves threads to other category and syncs both categories"""
        response = self.patch(
            self.api_link,
            {
                "ids": self.ids,
                "ops": [
                    {
                        "op": "replace",
                        "path": "category",
                        "value": self.other_category.id,
                    },
                    {"op": "replace", "path": "flatten-categories", "value": None},
                ],
            },
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        for i, thread in enumerate(self.threads):
            self.assertEqual(response_json[i]["id"], thread.id)
            self.assertEqual(response_json[i]["category"], self.other_category.id)

        for thread in Thread.objects.filter(id__in=self.ids):
            self.assertEqual(thread.category_id, self.other_category.id)

        category = Category.objects.get(pk=self.category.id)
        self.assertEqual(category.threads, self.category.threads - 3)

        new_category = Category.objects.get(pk=self.other_category.id)
        self.assertEqual(new_category.threads, 3)


class BulkThreadsHideApiTests(ThreadsBulkPatchApiTestCase):
    @patch_category_acl({"can_hide_threads": 1})
    def test_hide_thread(self):
        """api makes it possible to hide thread"""
        response = self.patch(
            self.api_link,
            {
                "ids": self.ids,
                "ops": [{"op": "replace", "path": "is-hidden", "value": True}],
            },
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        for i, thread in enumerate(self.threads):
            self.assertEqual(response_json[i]["id"], thread.id)
            self.assertTrue(response_json[i]["is_hidden"])

        for thread in Thread.objects.filter(id__in=self.ids):
            self.assertTrue(thread.is_hidden)

        category = Category.objects.get(pk=self.category.id)
        self.assertNotIn(category.last_thread_id, self.ids)


class BulkThreadsApproveApiTests(ThreadsBulkPatchApiTestCase):
    @patch_category_acl({"can_approve_content": True})
    def test_approve_thread(self):
        """api approvse threads and syncs category"""
        for thread in self.threads:
            thread.first_post.is_unapproved = True
            thread.first_post.save()

            thread.synchronize()
            thread.save()

            self.assertTrue(thread.is_unapproved)
            self.assertTrue(thread.has_unapproved_posts)

        self.category.synchronize()
        self.category.save()

        response = self.patch(
            self.api_link,
            {
                "ids": self.ids,
                "ops": [{"op": "replace", "path": "is-unapproved", "value": False}],
            },
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        for i, thread in enumerate(self.threads):
            self.assertEqual(response_json[i]["id"], thread.id)
            self.assertFalse(response_json[i]["is_unapproved"])
            self.assertFalse(response_json[i]["has_unapproved_posts"])

        for thread in Thread.objects.filter(id__in=self.ids):
            self.assertFalse(thread.is_unapproved)
            self.assertFalse(thread.has_unapproved_posts)

        category = Category.objects.get(pk=self.category.id)
        self.assertIn(category.last_thread_id, self.ids)



from django.contrib.auth import get_user_model
from django.shortcuts import redirect
from django.templatetags.static import static

from ...conf import settings

User = get_user_model()


def user_avatar(request, pk, size):
    size = int(size)

    try:
        user = User.objects.get(pk=pk)
    except User.DoesNotExist:
        return blank_avatar(request)

    found_avatar = user.avatars[0]
    for avatar in user.avatars:
        if avatar["size"] >= size:
            found_avatar = avatar
    return redirect(found_avatar["url"])


def blank_avatar(request):
    return redirect(
        request.settings.blank_avatar or static(settings.MISAGO_BLANK_AVATAR)
    )



import pytest

from ..exceptions import NotificationVerbError
from ..registry import NotificationRegistry, registry
from ..models import Notification
from ..verbs import NotificationVerb


@pytest.fixture
def request_mock(rf, user):
    request = rf.get("/notification/1/")

    request.user = user

    return request


def test_notification_registry_can_have_message_set_with_setter():
    notification_registry = NotificationRegistry()
    notification_registry.message("TEST", lambda obj: f"Hello #{obj.id}")

    message = notification_registry.get_message(Notification(id=1, verb="TEST"))
    assert message == "Hello #1"


def test_notification_registry_can_have_message_set_with_decorator():
    notification_registry = NotificationRegistry()

    @notification_registry.message("TEST")
    def get_test_message(obj):
        return f"Hello #{obj.id}"

    message = notification_registry.get_message(Notification(id=1, verb="TEST"))
    assert message == "Hello #1"


def test_notification_registry_get_message_raises_verb_error_for_unknown_verb():
    notification_registry = NotificationRegistry()

    with pytest.raises(NotificationVerbError) as excinfo:
        notification_registry.get_message(Notification(id=1, verb="TEST"))

    assert "TEST" in str(excinfo)


def test_notification_registry_can_have_redirect_set_with_setter(request_mock):
    notification_registry = NotificationRegistry()
    notification_registry.redirect("TEST", lambda _, obj: f"/test/#{obj.id}")

    redirect = notification_registry.get_redirect_url(
        request_mock, Notification(id=1, verb="TEST")
    )
    assert redirect == "/test/#1"


def test_notification_registry_can_have_redirect_set_with_decorator(
    request_mock,
):
    notification_registry = NotificationRegistry()

    @notification_registry.redirect("TEST")
    def get_test_redirect_url(request, obj):
        return f"/test/#{obj.id}"

    redirect = notification_registry.get_redirect_url(
        request_mock, Notification(id=1, verb="TEST")
    )
    assert redirect == "/test/#1"


def test_notification_registry_get_redirect_url_raises_verb_error_for_unknown_verb(
    request_mock,
):
    notification_registry = NotificationRegistry()

    with pytest.raises(NotificationVerbError) as excinfo:
        notification_registry.get_redirect_url(
            request_mock, Notification(id=1, verb="TEST")
        )

    assert "TEST" in str(excinfo)


def test_default_notification_registry_supports_test_notifications(request_mock):
    message = registry.get_message(Notification(id=1, verb="TEST"))
    assert message == "Test notification #1"

    redirect = registry.get_redirect_url(request_mock, Notification(id=1, verb="TEST"))
    assert redirect == "/#test-notification-1"


def test_default_notification_registry_supports_reply_notifications():
    message = registry.get_message(
        Notification(
            id=1,
            verb=NotificationVerb.REPLIED,
            actor_name="Aerith",
            thread_title="Midgar was destroyed!",
        )
    )
    assert message == ("<b>Aerith</b> replied to <b>Midgar was destroyed!</b>")


def test_default_notification_registry_supports_invite_notifications():
    message = registry.get_message(
        Notification(
            id=1,
            verb=NotificationVerb.INVITED,
            actor_name="Aerith",
            thread_title="Midgar was destroyed!",
        )
    )
    assert message == ("<b>Aerith</b> invited you to <b>Midgar was destroyed!</b>")



from django import forms
from django.contrib.auth import authenticate, get_user_model
from django.contrib.auth.forms import AuthenticationForm as BaseAuthenticationForm
from django.core.exceptions import ValidationError
from django.core.validators import validate_email
from django.utils.translation import pgettext, pgettext_lazy

from ..bans import get_user_ban

User = get_user_model()


class MisagoAuthMixin:
    error_messages = {
        "empty_data": pgettext_lazy("auth form", "Fill out both fields."),
        "invalid_login": pgettext_lazy("auth form", "Login or password is incorrect."),
        "inactive_user": pgettext_lazy(
            "auth form",
            "You have to activate your account before you will be able to sign in.",
        ),
        "inactive_admin": pgettext_lazy(
            "auth form",
            "Your account has to be activated by site administrator before you will be able to sign in.",
        ),
    }

    def confirm_user_active(self, user):
        if user.requires_activation_by_admin:
            raise ValidationError(
                self.error_messages["inactive_admin"], code="inactive_admin"
            )

        if user.requires_activation_by_user:
            raise ValidationError(
                self.error_messages["inactive_user"], code="inactive_user"
            )

    def confirm_user_not_banned(self, user):
        if not user.is_staff:
            self.user_ban = get_user_ban(user, self.request.cache_versions)
            if self.user_ban:
                raise ValidationError("", code="banned")

    def get_errors_dict(self):
        error = self.errors.as_data()["__all__"][0]
        if error.code == "banned":
            error.message = self.user_ban.ban.get_serialized_message()
        else:
            error.message = error.messages[0]

        return {"detail": error.message, "code": error.code}


class AuthenticationForm(MisagoAuthMixin, BaseAuthenticationForm):
    """
    Base class for authenticating users, Floppy-forms and
    Misago login field compliant
    """

    username = forms.CharField(
        label=pgettext_lazy("login form", "Username or e-mail"),
        required=False,
        max_length=254,
    )
    password = forms.CharField(
        label=pgettext_lazy("login form", "Password"),
        strip=False,
        required=False,
        widget=forms.PasswordInput,
    )

    def __init__(self, *args, request=None, **kwargs):
        self.request = request
        super().__init__(*args, **kwargs)

    def clean(self):
        username = self.cleaned_data.get("username")
        password = self.cleaned_data.get("password")

        if username and password:
            self.user_cache = authenticate(username=username, password=password)

            if self.user_cache is None or not self.user_cache.is_active:
                raise ValidationError(
                    self.error_messages["invalid_login"], code="invalid_login"
                )
            else:
                self.confirm_login_allowed(self.user_cache)
        else:
            raise ValidationError(self.error_messages["empty_data"], code="empty_data")

        return self.cleaned_data

    def confirm_login_allowed(self, user):
        self.confirm_user_active(user)
        self.confirm_user_not_banned(user)


class AdminAuthenticationForm(AuthenticationForm):
    required_css_class = "required"

    def __init__(self, *args, **kwargs):
        self.error_messages.update(
            {
                "not_staff": pgettext_lazy(
                    "admin login form", "Your account does not have admin privileges."
                )
            }
        )

        super().__init__(*args, **kwargs)

    def confirm_login_allowed(self, user):
        if not user.is_staff:
            raise forms.ValidationError(
                self.error_messages["not_staff"], code="not_staff"
            )


class GetUserForm(MisagoAuthMixin, forms.Form):
    email = forms.CharField()

    def clean(self):
        data = super().clean()

        email = data.get("email")
        if not email or len(email) > 250:
            raise forms.ValidationError(
                pgettext_lazy("auth email form", "Enter e-mail address."),
                code="empty_email",
            )

        try:
            validate_email(email)
        except forms.ValidationError:
            raise forms.ValidationError(
                pgettext_lazy("auth email form", "Entered e-mail is invalid."),
                code="invalid_email",
            )

        try:
            user = User.objects.get_by_email(data["email"])
            if not user.is_active:
                raise User.DoesNotExist()
            self.user_cache = user
        except User.DoesNotExist:
            raise forms.ValidationError(
                pgettext_lazy("auth email form", "No user with this e-mail exists."),
                code="not_found",
            )

        self.confirm_allowed(user)

        return data

    def confirm_allowed(self, user):
        """override this method to include additional checks"""


class ResendActivationForm(GetUserForm):
    def confirm_allowed(self, user):
        username_format = {"user": user.username}

        if not user.requires_activation:
            message = pgettext(
                "resend activation form", "%(user)s, your account is already active."
            )
            raise forms.ValidationError(
                message % username_format, code="already_active"
            )

        if user.requires_activation_by_admin:
            message = pgettext(
                "resend activation form",
                "%(user)s, only administrator may activate your account.",
            )
            raise forms.ValidationError(
                message % username_format, code="inactive_admin"
            )


class ResetPasswordForm(GetUserForm):
    error_messages = {
        "inactive_user": pgettext_lazy(
            "reset password form",
            "You have to activate your account before you will be able to request new password.",
        ),
        "inactive_admin": pgettext_lazy(
            "reset password form",
            "Administrator has to activate your account before you will be able to request new password.",
        ),
    }

    def confirm_allowed(self, user):
        self.confirm_user_active(user)



from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.utils.translation import pgettext
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.response import Response

from ...categories import PRIVATE_THREADS_ROOT_NAME, THREADS_ROOT_NAME
from ...core.shortcuts import get_int_or_404
from ..models import Post, Thread
from ..permissions import allow_use_private_threads
from ..viewmodels import (
    ForumThread,
    PrivateThread,
    PrivateThreadsCategory,
    ThreadsRootCategory,
)
from .postingendpoint import PostingEndpoint
from .threadendpoints.delete import delete_bulk, delete_thread
from .threadendpoints.editor import thread_start_editor
from .threadendpoints.list import private_threads_list_endpoint, threads_list_endpoint
from .threadendpoints.merge import thread_merge_endpoint, threads_merge_endpoint
from .threadendpoints.patch import bulk_patch_endpoint, thread_patch_endpoint


class ViewSet(viewsets.ViewSet):
    thread = None

    def get_thread(
        self, request, pk, path_aware=False, read_aware=False, watch_aware=False
    ):
        return self.thread(  # pylint: disable=not-callable
            request,
            get_int_or_404(pk),
            path_aware=path_aware,
            read_aware=read_aware,
            watch_aware=watch_aware,
        )

    def retrieve(self, request, pk):
        thread = self.get_thread(
            request, pk, path_aware=True, read_aware=True, watch_aware=True
        )

        return Response(thread.get_frontend_context())

    @transaction.atomic
    def partial_update(self, request, pk=None):
        thread = self.get_thread(request, pk).unwrap()
        return thread_patch_endpoint(request, thread)

    def patch(self, request):
        return bulk_patch_endpoint(request, self.thread)

    def delete(self, request, pk=None):
        if pk:
            thread = self.get_thread(request, pk).unwrap()
            return delete_thread(request, thread)
        return delete_bulk(request, self.thread)


class ThreadViewSet(ViewSet):
    category = ThreadsRootCategory
    thread = ForumThread

    def list(self, request):
        return threads_list_endpoint(request)

    @transaction.atomic
    def create(self, request):
        # Initialize empty instances for new thread
        thread = Thread()
        post = Post(thread=thread)

        # Put them through posting pipeline
        posting = PostingEndpoint(
            request,
            PostingEndpoint.START,
            tree_name=THREADS_ROOT_NAME,
            thread=thread,
            post=post,
        )

        if not posting.is_valid():
            return Response(posting.errors, status=400)

        posting.save()

        return Response(
            {"id": thread.pk, "title": thread.title, "url": thread.get_absolute_url()}
        )

    @action(detail=True, methods=["post"], url_path="merge", url_name="merge")
    @transaction.atomic
    def thread_merge(self, request, pk=None):
        thread = self.get_thread(request, pk).unwrap()
        return thread_merge_endpoint(request, thread, self.thread)

    @action(detail=False, methods=["post"], url_path="merge", url_name="merge")
    @transaction.atomic
    def threads_merge(self, request):
        return threads_merge_endpoint(request)

    @action(detail=False, methods=["get"])
    def editor(self, request):
        return thread_start_editor(request)


class PrivateThreadViewSet(ViewSet):
    category = PrivateThreadsCategory
    thread = PrivateThread

    def list(self, request):
        return private_threads_list_endpoint(request)

    @transaction.atomic
    def create(self, request):
        allow_use_private_threads(request.user_acl)
        if not request.user_acl["can_start_private_threads"]:
            raise PermissionDenied(
                pgettext("private threads api", "You can't start private threads.")
            )

        request.user.lock()

        # Initialize empty instances for new thread
        thread = Thread()
        post = Post(thread=thread)

        # Put them through posting pipeline
        posting = PostingEndpoint(
            request,
            PostingEndpoint.START,
            tree_name=PRIVATE_THREADS_ROOT_NAME,
            thread=thread,
            post=post,
        )

        if not posting.is_valid():
            return Response(posting.errors, status=400)

        posting.save()

        return Response(
            {"id": thread.pk, "title": thread.title, "url": thread.get_absolute_url()}
        )



from django import forms
from django.utils.translation import pgettext_lazy

from ....notifications.threads import ThreadNotifications
from .base import SettingsForm


class NotificationsSettingsForm(SettingsForm):
    settings = [
        "watch_started_threads",
        "watch_replied_threads",
        "watch_new_private_threads_by_followed",
        "watch_new_private_threads_by_other_users",
        "notify_new_private_threads_by_followed",
        "notify_new_private_threads_by_other_users",
        "delete_notifications_older_than",
    ]

    watch_started_threads = forms.TypedChoiceField(
        label=pgettext_lazy(
            "admin notifications settings form",
            "Notify about new replies in threads started by them",
        ),
        choices=ThreadNotifications.choices,
        widget=forms.RadioSelect(),
        coerce=int,
    )
    watch_replied_threads = forms.TypedChoiceField(
        label=pgettext_lazy(
            "admin notifications settings form",
            "Notify about new replies in threads replied to by them",
        ),
        choices=ThreadNotifications.choices,
        widget=forms.RadioSelect(),
        coerce=int,
    )

    watch_new_private_threads_by_followed = forms.TypedChoiceField(
        label=pgettext_lazy(
            "admin notifications settings form",
            "Notify about new replies in new private threads started by followed users",
        ),
        choices=ThreadNotifications.choices,
        widget=forms.RadioSelect(),
        coerce=int,
    )
    watch_new_private_threads_by_other_users = forms.TypedChoiceField(
        label=pgettext_lazy(
            "admin notifications settings form",
            "Notify about new replies in new private threads started by other users",
        ),
        choices=ThreadNotifications.choices,
        widget=forms.RadioSelect(),
        coerce=int,
    )

    notify_new_private_threads_by_followed = forms.TypedChoiceField(
        label=pgettext_lazy(
            "admin notifications settings form",
            "Notify about new private threads started by followed users",
        ),
        choices=ThreadNotifications.choices,
        widget=forms.RadioSelect(),
        coerce=int,
    )
    notify_new_private_threads_by_other_users = forms.TypedChoiceField(
        label=pgettext_lazy(
            "admin notifications settings form",
            "Notify about new private threads started by other users",
        ),
        choices=ThreadNotifications.choices,
        widget=forms.RadioSelect(),
        coerce=int,
    )

    delete_notifications_older_than = forms.IntegerField(
        label=pgettext_lazy(
            "admin notifications settings form",
            "Delete notifications older than (in days)",
        ),
        help_text=pgettext_lazy(
            "admin notifications settings form",
            "Misago automatically deletes notifications older than the specified age.",
        ),
        min_value=1,
    )



import pytest

from .. import SETTINGS_CACHE
from ..dynamicsettings import DynamicSettings


def test_settings_are_loaded_from_database_if_cache_is_not_available(
    db, mocker, cache_versions, django_assert_num_queries
):
    mocker.patch("django.core.cache.cache.get", return_value=None)
    with django_assert_num_queries(1):
        DynamicSettings(cache_versions)


def test_settings_are_loaded_from_cache_if_it_is_set(
    db, mocker, cache_versions, django_assert_num_queries
):
    cache_get = mocker.patch("django.core.cache.cache.get", return_value={})
    with django_assert_num_queries(0):
        DynamicSettings(cache_versions)
    cache_get.assert_called_once()


def test_settings_cache_is_set_if_none_exists(db, mocker, cache_versions):
    cache_set = mocker.patch("django.core.cache.cache.set")
    mocker.patch("django.core.cache.cache.get", return_value=None)

    DynamicSettings(cache_versions)
    cache_set.assert_called_once()


def test_settings_cache_is_not_set_if_it_already_exists(
    db, mocker, cache_versions, django_assert_num_queries
):
    cache_set = mocker.patch("django.core.cache.cache.set")
    mocker.patch("django.core.cache.cache.get", return_value={})
    with django_assert_num_queries(0):
        DynamicSettings(cache_versions)
    cache_set.assert_not_called()


def test_settings_cache_key_includes_cache_name_and_version(db, mocker, cache_versions):
    cache_set = mocker.patch("django.core.cache.cache.set")
    mocker.patch("django.core.cache.cache.get", return_value=None)
    DynamicSettings(cache_versions)
    cache_key = cache_set.call_args[0][0]
    assert SETTINGS_CACHE in cache_key
    assert cache_versions[SETTINGS_CACHE] in cache_key


def test_accessing_attr_returns_setting_value(db, cache_versions):
    settings = DynamicSettings(cache_versions)
    assert settings.forum_name == "Misago"


def test_accessing_attr_for_undefined_setting_raises_attribute_error(
    db, cache_versions
):
    settings = DynamicSettings(cache_versions)
    with pytest.raises(AttributeError):
        settings.not_existing  # pylint: disable=pointless-statement


def test_accessing_attr_for_lazy_setting_with_value_returns_true(
    cache_versions, lazy_setting
):
    settings = DynamicSettings(cache_versions)
    assert settings.lazy_setting is True


def test_lazy_setting_getter_for_lazy_setting_with_value_returns_real_value(
    cache_versions, lazy_setting
):
    settings = DynamicSettings(cache_versions)
    assert settings.get_lazy_setting_value("lazy_setting") == lazy_setting.value


def test_lazy_setting_getter_for_lazy_setting_makes_db_query(
    cache_versions, lazy_setting, django_assert_num_queries
):
    settings = DynamicSettings(cache_versions)
    with django_assert_num_queries(1):
        settings.get_lazy_setting_value("lazy_setting")


def test_accessing_attr_for_lazy_setting_without_value_returns_none(
    cache_versions, lazy_setting_without_value
):
    settings = DynamicSettings(cache_versions)
    assert settings.lazy_setting is None


def test_lazy_setting_getter_for_lazy_setting_is_reusing_query_result(
    cache_versions, lazy_setting, django_assert_num_queries
):
    settings = DynamicSettings(cache_versions)
    settings.get_lazy_setting_value("lazy_setting")
    with django_assert_num_queries(0):
        settings.get_lazy_setting_value("lazy_setting")


def test_lazy_setting_getter_for_undefined_setting_raises_attribute_error(
    db, cache_versions
):
    settings = DynamicSettings(cache_versions)
    with pytest.raises(AttributeError):
        settings.get_lazy_setting_value("undefined")


def test_lazy_setting_getter_for_not_lazy_setting_raises_value_error(
    db, cache_versions
):
    settings = DynamicSettings(cache_versions)
    with pytest.raises(ValueError):
        settings.get_lazy_setting_value("forum_name")


def test_public_settings_getter_returns_dict_with_public_settings(
    cache_versions, public_setting
):
    settings = DynamicSettings(cache_versions)
    public_settings = settings.get_public_settings()
    assert public_settings["public_setting"] == "Hello"


def test_public_settings_getter_excludes_private_settings_from_dict(
    cache_versions, private_setting
):
    settings = DynamicSettings(cache_versions)
    public_settings = settings.get_public_settings()
    assert "private_setting" not in public_settings


def test_getter_returns_setting_dict(cache_versions, public_setting):
    settings = DynamicSettings(cache_versions)
    assert settings.get(public_setting.setting) == {
        "value": public_setting.value,
        "is_lazy": public_setting.is_lazy,
        "is_public": public_setting.is_public,
        "width": public_setting.image_width,
        "height": public_setting.image_height,
    }



from .cache import set_menus_cache, get_menus_cache
from .models import MenuItem


def get_navbar_menu_items(cache_versions):
    return get_items(cache_versions).get(MenuItem.MENU_NAVBAR)


def get_footer_menu_items(cache_versions):
    return get_items(cache_versions).get(MenuItem.MENU_FOOTER)


def get_items(cache_versions):
    items = get_menus_cache(cache_versions)
    if items is None:
        items = get_items_from_db()
        set_menus_cache(cache_versions, items)
    return items


def get_items_from_db():
    return {
        MenuItem.MENU_NAVBAR: get_navbar_menu_items_from_db(),
        MenuItem.MENU_FOOTER: get_footer_menu_items_from_db(),
    }


def get_navbar_menu_items_from_db():
    return MenuItem.objects.exclude(menu=MenuItem.MENU_FOOTER).values()


def get_footer_menu_items_from_db():
    return MenuItem.objects.exclude(menu=MenuItem.MENU_NAVBAR).values()



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy

from .pages import user_profile, usercp, users_list


class MisagoUsersConfig(AppConfig):
    name = "misago.users"
    label = "misago_users"
    verbose_name = "Misago Auth"

    def ready(self):
        from . import signals as _
        from .admin import tasks  # pylint: disable=unused-import

        self.register_default_usercp_pages()
        self.register_default_users_list_pages()
        self.register_default_user_profile_pages()

    def register_default_usercp_pages(self):
        def auth_is_not_delegated(request):
            return not request.settings.enable_oauth2_client

        usercp.add_section(
            link="misago:usercp-change-forum-options",
            name=pgettext_lazy("user options page", "Forum options"),
            component="forum-options",
            icon="settings",
        )
        usercp.add_section(
            link="misago:usercp-edit-details",
            name=pgettext_lazy("user options page", "Edit details"),
            component="edit-details",
            icon="person_outline",
        )
        usercp.add_section(
            link="misago:usercp-change-username",
            name=pgettext_lazy("user options page", "Change username"),
            component="change-username",
            icon="card_membership",
            visible_if=auth_is_not_delegated,
        )
        usercp.add_section(
            link="misago:usercp-change-email-password",
            name=pgettext_lazy("user options page", "Change e-mail or password"),
            component="sign-in-credentials",
            icon="vpn_key",
            visible_if=auth_is_not_delegated,
        )

        def can_download_own_data(request):
            return request.settings.allow_data_downloads

        usercp.add_section(
            link="misago:usercp-download-data",
            name=pgettext_lazy("user options page", "Download data"),
            component="download-data",
            icon="save_alt",
            visible_if=can_download_own_data,
        )

        def can_delete_own_account(request):
            if not auth_is_not_delegated(request):
                return False

            return request.settings.allow_delete_own_account

        usercp.add_section(
            link="misago:usercp-delete-account",
            name=pgettext_lazy("user options page", "Delete account"),
            component="delete-account",
            icon="cancel",
            visible_if=can_delete_own_account,
        )

    def register_default_users_list_pages(self):
        users_list.add_section(
            link="misago:users-active-posters",
            component="active-posters",
            name=pgettext_lazy("users lists page", "Top posters"),
        )

    def register_default_user_profile_pages(self):
        def can_see_names_history(request, profile):
            if request.user.is_authenticated:
                is_account_owner = profile.pk == request.user.pk
                has_permission = request.user_acl["can_see_users_name_history"]
                return is_account_owner or has_permission
            return False

        def can_see_ban_details(request, profile):
            if request.user.is_authenticated:
                if request.user_acl["can_see_ban_details"]:
                    from .bans import get_user_ban

                    return bool(get_user_ban(profile, request.cache_versions))
                return False
            return False

        user_profile.add_section(
            link="misago:user-posts",
            name=pgettext_lazy("user profile page", "Posts"),
            icon="message",
            component="posts",
        )
        user_profile.add_section(
            link="misago:user-threads",
            name=pgettext_lazy("user profile page", "Threads"),
            icon="forum",
            component="threads",
        )
        user_profile.add_section(
            link="misago:user-followers",
            name=pgettext_lazy("user profile page", "Followers"),
            icon="favorite",
            component="followers",
        )
        user_profile.add_section(
            link="misago:user-follows",
            name=pgettext_lazy("user profile page", "Follows"),
            icon="favorite_border",
            component="follows",
        )
        user_profile.add_section(
            link="misago:user-details",
            name=pgettext_lazy("user profile page", "Details"),
            icon="person_outline",
            component="details",
        )
        user_profile.add_section(
            link="misago:username-history",
            name=pgettext_lazy("user profile page", "Username history"),
            icon="card_membership",
            component="username-history",
            visible_if=can_see_names_history,
        )
        user_profile.add_section(
            link="misago:user-ban",
            name=pgettext_lazy("user profile page", "Ban details"),
            icon="remove_circle_outline",
            component="ban-details",
            visible_if=can_see_ban_details,
        )



from typing import TYPE_CHECKING

from django.core.exceptions import PermissionDenied
from django.utils.translation import pgettext

if TYPE_CHECKING:
    from misago.users.models import User


def allow_use_notifications(user: "User"):
    if user.is_anonymous:
        raise PermissionDenied(
            pgettext(
                "notifications permission",
                "You must be signed in to access your notifications.",
            )
        )



from datetime import timedelta

from django.test import TestCase
from django.utils import timezone

from ...categories.models import Category
from ...users.test import create_test_user
from ..checksums import update_post_checksum
from ..models import Post, Thread


class PostModelTests(TestCase):
    def setUp(self):
        self.user = create_test_user("User", "user@example.com")

        datetime = timezone.now()

        self.category = Category.objects.all_categories()[:1][0]
        self.thread = Thread(
            category=self.category,
            started_on=datetime,
            starter_name="Tester",
            starter_slug="tester",
            last_post_on=datetime,
            last_poster_name="Tester",
            last_poster_slug="tester",
        )

        self.thread.set_title("Test thread")
        self.thread.save()

        self.post = Post.objects.create(
            category=self.category,
            thread=self.thread,
            poster=self.user,
            poster_name=self.user.username,
            original="Hello! I am test message!",
            parsed="<p>Hello! I am test message!</p>",
            checksum="nope",
            posted_on=datetime,
            updated_on=datetime,
        )

        update_post_checksum(self.post)
        self.post.save(update_fields=["checksum"])

        self.thread.first_post = self.post
        self.thread.last_post = self.post
        self.thread.save()

    def test_merge_invalid(self):
        """see if attempts for invalid merges fail"""
        # can't merge with itself
        with self.assertRaises(ValueError):
            self.post.merge(self.post)

        other_user = create_test_user("Other_User", "otheruser@example.com")

        other_thread = Thread.objects.create(
            category=self.category,
            started_on=timezone.now(),
            starter_name="Tester",
            starter_slug="tester",
            last_post_on=timezone.now(),
            last_poster_name="Tester",
            last_poster_slug="tester",
        )

        # can't merge with other users posts
        with self.assertRaises(ValueError):
            self.post.merge(
                Post.objects.create(
                    category=self.category,
                    thread=self.thread,
                    poster=other_user,
                    poster_name=other_user.username,
                    original="Hello! I am test message!",
                    parsed="<p>Hello! I am test message!</p>",
                    checksum="nope",
                    posted_on=timezone.now() + timedelta(minutes=5),
                    updated_on=timezone.now() + timedelta(minutes=5),
                )
            )

        # can't merge across threads
        with self.assertRaises(ValueError):
            self.post.merge(
                Post.objects.create(
                    category=self.category,
                    thread=other_thread,
                    poster=self.user,
                    poster_name=self.user.username,
                    original="Hello! I am test message!",
                    parsed="<p>Hello! I am test message!</p>",
                    checksum="nope",
                    posted_on=timezone.now() + timedelta(minutes=5),
                    updated_on=timezone.now() + timedelta(minutes=5),
                )
            )

        # can't merge with events
        with self.assertRaises(ValueError):
            self.post.merge(
                Post.objects.create(
                    category=self.category,
                    thread=self.thread,
                    poster=self.user,
                    poster_name=self.user.username,
                    original="Hello! I am test message!",
                    parsed="<p>Hello! I am test message!</p>",
                    checksum="nope",
                    posted_on=timezone.now() + timedelta(minutes=5),
                    updated_on=timezone.now() + timedelta(minutes=5),
                    is_event=True,
                )
            )

    def test_merge(self):
        """merge method merges two posts into one"""
        other_post = Post.objects.create(
            category=self.category,
            thread=self.thread,
            poster=self.user,
            poster_name=self.user.username,
            original="I am other message!",
            parsed="<p>I am other message!</p>",
            checksum="nope",
            posted_on=timezone.now() + timedelta(minutes=5),
            updated_on=timezone.now() + timedelta(minutes=5),
        )

        other_post.merge(self.post)

        self.assertIn(other_post.original, self.post.original)
        self.assertIn(other_post.parsed, self.post.parsed)
        self.assertTrue(self.post.is_valid)

    def test_merge_best_answer(self):
        """merge method merges best answer into post"""
        best_answer = Post.objects.create(
            category=self.category,
            thread=self.thread,
            poster=self.user,
            poster_name=self.user.username,
            original="I am other message!",
            parsed="<p>I am other message!</p>",
            checksum="nope",
            posted_on=timezone.now() + timedelta(minutes=5),
            updated_on=timezone.now() + timedelta(minutes=5),
        )

        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        best_answer.merge(self.post)
        self.assertEqual(self.thread.best_answer, self.post)

    def test_merge_in_best_answer(self):
        """merge method merges post into best answert"""
        best_answer = Post.objects.create(
            category=self.category,
            thread=self.thread,
            poster=self.user,
            poster_name=self.user.username,
            original="I am other message!",
            parsed="<p>I am other message!</p>",
            checksum="nope",
            posted_on=timezone.now() + timedelta(minutes=5),
            updated_on=timezone.now() + timedelta(minutes=5),
        )

        other_post = Post.objects.create(
            category=self.category,
            thread=self.thread,
            poster=self.user,
            poster_name=self.user.username,
            original="I am other message!",
            parsed="<p>I am other message!</p>",
            checksum="nope",
            posted_on=timezone.now() + timedelta(minutes=5),
            updated_on=timezone.now() + timedelta(minutes=5),
        )

        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        other_post.merge(best_answer)
        self.assertEqual(self.thread.best_answer, best_answer)

    def test_move(self):
        """move method moves post to other thread"""
        new_thread = Thread.objects.create(
            category=self.category,
            started_on=timezone.now(),
            starter_name="Tester",
            starter_slug="tester",
            last_post_on=timezone.now(),
            last_poster_name="Tester",
            last_poster_slug="tester",
        )

        self.post.move(new_thread)

        self.assertEqual(self.post.thread, new_thread)



from django.urls import reverse

from ...test import assert_has_message
from ..auth import is_admin_authorized

admin_logout_link = reverse("misago:admin:logout")
admin_link = reverse("misago:admin:index")
site_logout_link = reverse("misago:logout")


def test_admin_can_logout_from_admin_site_but_stay_logged(admin_client, superuser):
    response = admin_client.post(admin_logout_link)
    assert response.wsgi_request.user == superuser
    assert not is_admin_authorized(response.wsgi_request)


def test_admin_is_redirected_to_login_form_on_logout(admin_client, superuser):
    response = admin_client.post(admin_logout_link)
    assert response.status_code == 302
    assert response["location"] == admin_link


def test_admin_is_displayed_message_after_logout(admin_client, superuser):
    response = admin_client.post(admin_logout_link)
    assert_has_message(response, "Your admin session has been closed.")


def test_admin_can_logout_from_entire_site(admin_client):
    response = admin_client.post(site_logout_link)
    assert response.wsgi_request.user.is_anonymous
    assert not is_admin_authorized(response.wsgi_request)



from rest_framework import status
from rest_framework.decorators import api_view
from rest_framework.response import Response

from . import common_flavour, finalize_markup
from .serializers import MarkupSerializer


@api_view(["POST"])
def parse_markup(request):
    serializer = MarkupSerializer(
        data=request.data, context={"settings": request.settings}
    )
    if not serializer.is_valid():
        errors_list = list(serializer.errors.values())[0]
        return Response({"detail": errors_list[0]}, status=status.HTTP_400_BAD_REQUEST)

    parsing_result = common_flavour(
        request, request.user, serializer.data["post"], force_shva=True
    )
    finalized = finalize_markup(parsing_result["parsed_text"])

    return Response({"parsed": finalized})



from datetime import timedelta

from django.test.client import BOUNDARY, MULTIPART_CONTENT, encode_multipart
from django.urls import reverse
from django.utils import timezone

from .. import test
from ...acl.test import patch_user_acl
from ...categories.models import Category
from ...users.test import AuthenticatedUserTestCase
from ..models import Post, Thread
from ..test import patch_category_acl


class EditReplyTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")
        self.thread = test.post_thread(category=self.category)
        self.post = test.reply_thread(self.thread, poster=self.user)

        self.api_link = reverse(
            "misago:api:thread-post-detail",
            kwargs={"thread_pk": self.thread.pk, "pk": self.post.pk},
        )

    def put(self, url, data=None):
        content = encode_multipart(BOUNDARY, data or {})
        return self.client.put(url, content, content_type=MULTIPART_CONTENT)

    def test_cant_edit_reply_as_guest(self):
        """user has to be authenticated to be able to edit reply"""
        self.logout_user()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 403)

    def test_thread_visibility(self):
        """thread's visibility is validated"""
        with patch_category_acl({"can_see": False}):
            response = self.put(self.api_link)
            self.assertEqual(response.status_code, 404)

        with patch_category_acl({"can_browse": False}):
            response = self.put(self.api_link)
            self.assertEqual(response.status_code, 404)

        with patch_category_acl({"can_see_all_threads": False}):
            response = self.put(self.api_link)
            self.assertEqual(response.status_code, 404)

    @patch_category_acl({"can_edit_posts": 0})
    def test_cant_edit_reply(self):
        """permission to edit reply is validated"""
        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't edit posts in this category."}
        )

    @patch_category_acl({"can_edit_posts": 1})
    def test_cant_edit_other_user_reply(self):
        """permission to edit reply by other users is validated"""
        self.post.poster = None
        self.post.save()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't edit other users posts in this category."},
        )

    @patch_category_acl({"can_edit_posts": 1, "post_edit_time": 1})
    def test_edit_too_old(self):
        """permission to edit reply within timelimit is validated"""
        self.post.posted_on = timezone.now() - timedelta(minutes=5)
        self.post.save()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't edit posts that are older than 1 minute."},
        )

    @patch_category_acl({"can_edit_posts": 1, "can_close_threads": False})
    def test_closed_category_no_permission(self):
        """permssion to edit reply in closed category is validated"""
        self.category.is_closed = True
        self.category.save()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This category is closed. You can't edit posts in it."},
        )

    @patch_category_acl({"can_edit_posts": 1, "can_close_threads": True})
    def test_closed_category(self):
        """permssion to edit reply in closed category is validated"""
        self.category.is_closed = True
        self.category.save()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 400)

    @patch_category_acl({"can_edit_posts": 1, "can_close_threads": False})
    def test_closed_thread_no_permission(self):
        """permssion to edit reply in closed thread is validated"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This thread is closed. You can't edit posts in it."},
        )

    @patch_category_acl({"can_edit_posts": 1, "can_close_threads": True})
    def test_closed_thread(self):
        """permssion to edit reply in closed thread is validated"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 400)

    @patch_category_acl({"can_edit_posts": 1, "can_protect_posts": False})
    def test_protected_post_no_permission(self):
        """permssion to edit protected post is validated"""
        self.post.is_protected = True
        self.post.save()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This post is protected. You can't edit it."}
        )

    @patch_category_acl({"can_edit_posts": 1, "can_protect_posts": True})
    def test_protected_post_no(self):
        """permssion to edit protected post is validated"""
        self.post.is_protected = True
        self.post.save()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 400)

    @patch_category_acl({"can_edit_posts": 1})
    def test_empty_data(self):
        """no data sent handling has no showstoppers"""
        response = self.put(self.api_link, data={})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"post": ["You have to enter a message."]})

    @patch_category_acl({"can_edit_posts": 1})
    def test_invalid_data(self):
        """api errors for invalid request data"""
        response = self.client.put(
            self.api_link, "false", content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "non_field_errors": [
                    "Invalid data. Expected a dictionary, but got bool."
                ]
            },
        )

    @patch_category_acl({"can_edit_posts": 1})
    def test_edit_event(self):
        """events can't be edited"""
        self.post.is_event = True
        self.post.save()

        response = self.put(self.api_link, data={})
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "Events can't be edited."})

    @patch_category_acl({"can_edit_posts": 1})
    def test_post_is_validated(self):
        """post is validated"""
        response = self.put(self.api_link, data={"post": "a"})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "post": [
                    "Posted message should be at least 5 characters long (it has 1)."
                ]
            },
        )

    @patch_category_acl({"can_edit_posts": 1})
    def test_edit_reply_no_change(self):
        """endpoint isn't bumping edits count if no change was made to post's body"""
        self.assertEqual(self.post.edits_record.count(), 0)

        response = self.put(self.api_link, data={"post": self.post.original})
        self.assertEqual(response.status_code, 200)

        response = self.client.get(self.thread.get_absolute_url())
        self.assertContains(response, self.post.parsed)

        post = self.thread.post_set.order_by("id").last()
        self.assertEqual(post.edits, 0)
        self.assertEqual(post.original, self.post.original)
        self.assertIsNone(post.last_editor_id, self.user.id)
        self.assertIsNone(post.last_editor_name, self.user.username)
        self.assertIsNone(post.last_editor_slug, self.user.slug)

        self.assertEqual(self.post.edits_record.count(), 0)

    @patch_category_acl({"can_edit_posts": 1})
    def test_edit_reply(self):
        """endpoint updates reply"""
        self.assertEqual(self.post.edits_record.count(), 0)

        response = self.put(self.api_link, data={"post": "This is test edit!"})
        self.assertEqual(response.status_code, 200)

        response = self.client.get(self.thread.get_absolute_url())
        self.assertContains(response, "<p>This is test edit!</p>")

        self.assertEqual(self.user.audittrail_set.count(), 1)

        post = self.thread.post_set.order_by("id").last()
        self.assertEqual(post.edits, 1)
        self.assertEqual(post.original, "This is test edit!")
        self.assertEqual(post.last_editor_id, self.user.id)
        self.assertEqual(post.last_editor_name, self.user.username)
        self.assertEqual(post.last_editor_slug, self.user.slug)

        self.assertEqual(self.post.edits_record.count(), 1)

        post_edit = post.edits_record.last()
        self.assertEqual(post_edit.edited_from, self.post.original)
        self.assertEqual(post_edit.edited_to, post.original)

        self.assertEqual(post_edit.editor_id, self.user.id)
        self.assertEqual(post_edit.editor_name, self.user.username)
        self.assertEqual(post_edit.editor_slug, self.user.slug)

    @patch_category_acl({"can_edit_posts": 2, "can_hide_threads": 1})
    def test_edit_first_post_hidden(self):
        """endpoint updates hidden thread's first post"""
        self.thread.is_hidden = True
        self.thread.save()
        self.thread.first_post.is_hidden = True
        self.thread.first_post.save()

        api_link = reverse(
            "misago:api:thread-post-detail",
            kwargs={"thread_pk": self.thread.pk, "pk": self.thread.first_post.pk},
        )

        response = self.put(api_link, data={"post": "This is test edit!"})
        self.assertEqual(response.status_code, 200)

    @patch_category_acl({"can_edit_posts": 1, "can_protect_posts": True})
    def test_protect_post(self):
        """can protect post"""
        response = self.put(
            self.api_link, data={"post": "Lorem ipsum dolor met!", "protect": 1}
        )
        self.assertEqual(response.status_code, 200)

        post = self.user.post_set.order_by("id").last()
        self.assertTrue(post.is_protected)

    @patch_category_acl({"can_edit_posts": 1, "can_protect_posts": False})
    def test_protect_post_no_permission(self):
        """cant protect post without permission"""
        response = self.put(
            self.api_link, data={"post": "Lorem ipsum dolor met!", "protect": 1}
        )
        self.assertEqual(response.status_code, 200)

        post = self.user.post_set.order_by("id").last()
        self.assertFalse(post.is_protected)

    @patch_category_acl({"can_edit_posts": 1})
    def test_post_unicode(self):
        """unicode characters can be posted"""
        response = self.put(
            self.api_link, data={"post": "Chrzążczyżewoszyce, powiat Łękółody."}
        )
        self.assertEqual(response.status_code, 200)

    @patch_category_acl({"can_edit_posts": 1})
    def test_reply_category_moderation_queue(self):
        """edit sends reply to queue due to category setup"""
        self.category.require_edits_approval = True
        self.category.save()

        response = self.put(self.api_link, data={"post": "Lorem ipsum dolor met!"})
        self.assertEqual(response.status_code, 200)

        post = self.user.post_set.all()[:1][0]
        self.assertTrue(post.is_unapproved)

    @patch_category_acl({"can_edit_posts": 1})
    @patch_user_acl({"can_approve_content": True})
    def test_reply_category_moderation_queue_bypass(self):
        """bypass moderation queue due to user's acl"""
        self.category.require_edits_approval = True
        self.category.save()

        response = self.put(self.api_link, data={"post": "Lorem ipsum dolor met!"})
        self.assertEqual(response.status_code, 200)

        post = self.user.post_set.all()[:1][0]
        self.assertFalse(post.is_unapproved)

    @patch_category_acl({"can_edit_posts": 1, "require_edits_approval": True})
    def test_reply_user_moderation_queue(self):
        """edit sends reply to queue due to user acl"""
        response = self.put(self.api_link, data={"post": "Lorem ipsum dolor met!"})
        self.assertEqual(response.status_code, 200)

        post = self.user.post_set.all()[:1][0]
        self.assertTrue(post.is_unapproved)

    @patch_category_acl({"can_edit_posts": 1, "require_edits_approval": True})
    @patch_user_acl({"can_approve_content": True})
    def test_reply_user_moderation_queue_bypass(self):
        """bypass moderation queue due to user's acl"""
        response = self.put(self.api_link, data={"post": "Lorem ipsum dolor met!"})
        self.assertEqual(response.status_code, 200)

        post = self.user.post_set.all()[:1][0]
        self.assertFalse(post.is_unapproved)

    @patch_category_acl(
        {
            "can_edit_posts": 1,
            "require_threads_approval": True,
            "require_replies_approval": True,
        }
    )
    def test_reply_omit_other_moderation_queues(self):
        """other queues are omitted"""
        self.category.require_threads_approval = True
        self.category.require_replies_approval = True
        self.category.save()

        response = self.put(self.api_link, data={"post": "Lorem ipsum dolor met!"})
        self.assertEqual(response.status_code, 200)

        post = self.user.post_set.all()[:1][0]
        self.assertFalse(post.is_unapproved)

    def setUpFirstReplyTest(self):
        self.post = self.thread.first_post

        self.post.poster = self.user
        self.post.save()

        self.api_link = reverse(
            "misago:api:thread-post-detail",
            kwargs={"thread_pk": self.thread.pk, "pk": self.post.pk},
        )

    @patch_category_acl({"can_edit_posts": 1})
    def test_first_reply_category_moderation_queue(self):
        """edit sends thread to queue due to category setup"""
        self.setUpFirstReplyTest()

        self.category.require_edits_approval = True
        self.category.save()

        response = self.put(self.api_link, data={"post": "Lorem ipsum dolor met!"})
        self.assertEqual(response.status_code, 200)

        thread = Thread.objects.get(pk=self.thread.pk)
        self.assertTrue(thread.is_unapproved)
        self.assertTrue(thread.has_unapproved_posts)

        post = Post.objects.get(pk=self.post.pk)
        self.assertTrue(post.is_unapproved)

    @patch_category_acl({"can_edit_posts": 1})
    @patch_user_acl({"can_approve_content": True})
    def test_first_reply_category_moderation_queue_bypass(self):
        """bypass moderation queue due to user's acl"""
        self.setUpFirstReplyTest()

        self.category.require_edits_approval = True
        self.category.save()

        response = self.put(self.api_link, data={"post": "Lorem ipsum dolor met!"})
        self.assertEqual(response.status_code, 200)

        thread = Thread.objects.get(pk=self.thread.pk)
        self.assertFalse(thread.is_unapproved)
        self.assertFalse(thread.has_unapproved_posts)

        post = Post.objects.get(pk=self.post.pk)
        self.assertFalse(post.is_unapproved)

    @patch_category_acl({"can_edit_posts": 1, "require_edits_approval": True})
    def test_first_reply_user_moderation_queue(self):
        """edit sends thread to queue due to user acl"""
        self.setUpFirstReplyTest()

        response = self.put(self.api_link, data={"post": "Lorem ipsum dolor met!"})
        self.assertEqual(response.status_code, 200)

        thread = Thread.objects.get(pk=self.thread.pk)
        self.assertTrue(thread.is_unapproved)
        self.assertTrue(thread.has_unapproved_posts)

        post = Post.objects.get(pk=self.post.pk)
        self.assertTrue(post.is_unapproved)

    @patch_category_acl({"can_edit_posts": 1, "require_edits_approval": True})
    @patch_user_acl({"can_approve_content": True})
    def test_first_reply_user_moderation_queue_bypass(self):
        """bypass moderation queue due to user's acl"""
        self.setUpFirstReplyTest()

        response = self.put(self.api_link, data={"post": "Lorem ipsum dolor met!"})
        self.assertEqual(response.status_code, 200)

        thread = Thread.objects.get(pk=self.thread.pk)
        self.assertFalse(thread.is_unapproved)
        self.assertFalse(thread.has_unapproved_posts)

        post = Post.objects.get(pk=self.post.pk)
        self.assertFalse(post.is_unapproved)

    @patch_category_acl(
        {
            "can_edit_posts": 1,
            "require_threads_approval": True,
            "require_replies_approval": True,
        }
    )
    def test_first_reply_omit_other_moderation_queues(self):
        """other queues are omitted"""
        self.setUpFirstReplyTest()

        self.category.require_threads_approval = True
        self.category.require_replies_approval = True
        self.category.save()

        response = self.put(self.api_link, data={"post": "Lorem ipsum dolor met!"})
        self.assertEqual(response.status_code, 200)

        thread = Thread.objects.get(pk=self.thread.pk)
        self.assertFalse(thread.is_unapproved)
        self.assertFalse(thread.has_unapproved_posts)

        post = Post.objects.get(pk=self.post.pk)
        self.assertFalse(post.is_unapproved)



from django.urls import reverse

from ...acl.test import patch_user_acl
from ...users.test import AuthenticatedUserTestCase
from ..models import Category


class CategoryViewsTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")

    def test_index_renders(self):
        """categories list renders for authenticated"""
        response = self.client.get(reverse("misago:categories"))
        self.assertContains(response, self.category.name)
        self.assertContains(response, self.category.get_absolute_url())

    def test_index_renders_for_guest(self):
        """categories list renders for guest"""
        self.logout_user()

        response = self.client.get(reverse("misago:categories"))
        self.assertContains(response, self.category.name)
        self.assertContains(response, self.category.get_absolute_url())

    @patch_user_acl({"visible_categories": []})
    def test_index_no_perms_renders(self):
        """categories list renders no visible categories for authenticated"""
        response = self.client.get(reverse("misago:categories"))
        self.assertNotContains(response, self.category.name)
        self.assertNotContains(response, self.category.get_absolute_url())

    @patch_user_acl({"visible_categories": []})
    def test_index_no_perms_renders_for_guest(self):
        """categories list renders no visible categories for guest"""
        self.logout_user()

        response = self.client.get(reverse("misago:categories"))
        self.assertNotContains(response, self.category.name)
        self.assertNotContains(response, self.category.get_absolute_url())



from unittest.mock import Mock

from django.core.exceptions import ValidationError
from django.test import TestCase

from ..models import Ban
from ..test import create_test_user
from ..validators import (
    validate_email,
    validate_email_available,
    validate_email_banned,
    validate_gmail_email,
    validate_username,
    validate_username_available,
    validate_username_banned,
    validate_username_content,
    validate_username_length,
)


class ValidateEmailAvailableTests(TestCase):
    def setUp(self):
        self.user = create_test_user("User", "user@example.com")

    def test_valid_email(self):
        """validate_email_available allows available emails"""
        validate_email_available("other@example.com")
        validate_email_available(self.user.email, exclude=self.user)

    def test_invalid_email(self):
        """validate_email_available disallows unvailable emails"""
        with self.assertRaises(ValidationError):
            validate_email_available(self.user.email)


class ValidateEmailBannedTests(TestCase):
    def setUp(self):
        Ban.objects.create(check_type=Ban.EMAIL, banned_value="ban@test.com")

    def test_unbanned_name(self):
        """unbanned email passes validation"""
        validate_email_banned("noban@test.com")

    def test_banned_name(self):
        """banned email fails validation"""
        with self.assertRaises(ValidationError):
            validate_email_banned("ban@test.com")


class ValidateEmailTests(TestCase):
    def test_validate_email(self):
        """validate_email has no crashes"""
        validate_email("user@example.com")
        with self.assertRaises(ValidationError):
            validate_email("*")


class ValidateUsernameTests(TestCase):
    def test_validate_username(self):
        """validate_username has no crashes"""
        settings = Mock(username_length_min=1, username_length_max=5)
        validate_username(settings, "LeBob")
        validate_username(settings, "123")
        validate_username(settings, "LeB_b")
        validate_username(settings, "_a_")
        validate_username(settings, "_123_")
        with self.assertRaises(ValidationError):
            validate_username(settings, "*")
        with self.assertRaises(ValidationError):
            validate_username(settings, "___")


class ValidateUsernameAvailableTests(TestCase):
    def setUp(self):
        self.user = create_test_user("User", "user@example.com")

    def test_valid_name(self):
        """validate_username_available allows available names"""
        validate_username_available("Other_User")
        validate_username_available(self.user.username, exclude=self.user)

    def test_invalid_name(self):
        """validate_username_available disallows unvailable names"""
        with self.assertRaises(ValidationError):
            validate_username_available(self.user.username)


class ValidateUsernameBannedTests(TestCase):
    def setUp(self):
        Ban.objects.create(check_type=Ban.USERNAME, banned_value="Ban")

    def test_unbanned_name(self):
        """unbanned name passes validation"""
        validate_username_banned("User")

    def test_banned_name(self):
        """banned name fails validation"""
        with self.assertRaises(ValidationError):
            validate_username_banned("Ban")


class ValidateUsernameContentTests(TestCase):
    def test_valid_name(self):
        """validate_username_content allows valid names"""
        validate_username_content("123")
        validate_username_content("User")
        validate_username_content("User123")

    def test_invalid_name(self):
        """validate_username_content disallows invalid names"""
        with self.assertRaises(ValidationError):
            validate_username_content("!")
        with self.assertRaises(ValidationError):
            validate_username_content("User!")
        with self.assertRaises(ValidationError):
            validate_username_content("John Doe")
        with self.assertRaises(ValidationError):
            validate_username_content("Rafał")
        with self.assertRaises(ValidationError):
            validate_username_content("初音 ミク")


class ValidateUsernameLengthTests(TestCase):
    def test_valid_name(self):
        """validate_username_length allows valid names"""
        settings = Mock(username_length_min=1, username_length_max=5)
        validate_username_length(settings, "a" * settings.username_length_min)
        validate_username_length(settings, "a" * settings.username_length_max)

    def test_invalid_name(self):
        """validate_username_length disallows invalid names"""
        settings = Mock(username_length_min=1, username_length_max=5)
        with self.assertRaises(ValidationError):
            validate_username_length(settings, "a" * (settings.username_length_min - 1))
        with self.assertRaises(ValidationError):
            validate_username_length(settings, "a" * (settings.username_length_max + 1))


class ValidateGmailEmailTests(TestCase):
    def test_validate_gmail_email(self):
        """validate_gmail_email spots spammy gmail address"""
        added_errors = {}

        def add_errors(field_name, errors):
            added_errors[field_name] = errors

        validate_gmail_email(None, {}, add_errors)
        validate_gmail_email(None, {"email": "invalid-email"}, add_errors)
        validate_gmail_email(None, {"email": "the.valid.email@gmail.com"}, add_errors)
        validate_gmail_email(None, {"email": "the.valid.email@hotmail.com"}, add_errors)

        self.assertFalse(added_errors)

        validate_gmail_email(
            None, {"email": "the.s.p.a.m.my.e.ma.il@gmail.com"}, add_errors
        )
        self.assertTrue(added_errors)



# Generated by Django 2.2.1 on 2019-05-18 16:31

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("misago_threads", "0010_auto_20180609_1523")]

    operations = [
        migrations.RunSQL("DROP INDEX IF EXISTS misago_thre_has_ope_479906_part"),
        migrations.RunSQL("DROP INDEX IF EXISTS misago_thre_is_hidd_85db69_part"),
        migrations.RunSQL("DROP INDEX IF EXISTS misago_thre_is_even_42bda7_part"),
        migrations.RunSQL("DROP INDEX IF EXISTS misago_thre_weight_955884_part"),
        migrations.RunSQL("DROP INDEX IF EXISTS misago_thre_weight_9e8f9c_part"),
        migrations.RunSQL("DROP INDEX IF EXISTS misago_thre_weight_c7ef29_part"),
        migrations.RunSQL("DROP INDEX IF EXISTS misago_thre_weight__4af9ee_part"),
        migrations.RunSQL("DROP INDEX IF EXISTS misago_thre_has_rep_84acfa_part"),
        migrations.RunSQL("DROP INDEX IF EXISTS misago_thre_has_una_b0dbf5_part"),
        migrations.RunSQL("DROP INDEX IF EXISTS misago_thre_is_hidd_d2b96c_part"),
    ]



import re

from django.forms import ValidationError
from django.utils.translation import pgettext

FILENAME_CONTENT = re.compile(r"([a-zA-Z0-9]|\.|_|-)+")
FILENAME_TANGIBILITY = re.compile(r"[a-zA-Z0-9]")


def validate_css_name(filename):
    if not filename.lower().endswith(".css"):
        raise ValidationError(
            pgettext("admin css name validator", "Name is missing an .css extension.")
        )

    if filename.startswith("."):
        raise ValidationError(
            pgettext("admin css name validator", 'Name can\'t start with period (".").')
        )

    if not FILENAME_CONTENT.fullmatch(filename[:-4]):
        raise ValidationError(
            pgettext(
                "admin css name validator",
                "Name can contain only latin alphabet characters, digits, dots, underscores and dashes.",
            )
        )

    if not FILENAME_TANGIBILITY.match(filename[:-4]):
        raise ValidationError(
            pgettext(
                "admin css name validator",
                "Name has to contain at least one latin alphabet character or digit.",
            )
        )


def validate_css_name_is_available(instance, name):
    queryset = instance.theme.css.filter(name=name)
    if instance.pk:
        queryset = queryset.exclude(pk=instance.pk)
    if queryset.exists():
        raise ValidationError(
            pgettext(
                "admin css name unique validator",
                "This name is already in use by other asset.",
            )
        )



from django.utils.translation import npgettext, pgettext
from rest_framework import serializers
from rest_framework.fields import empty

from . import PostingEndpoint, PostingMiddleware
from ....acl.objectacl import add_acl_to_obj
from ...serializers import AttachmentSerializer


class AttachmentsMiddleware(PostingMiddleware):
    def use_this_middleware(self):
        return bool(self.user_acl["max_attachment_size"])

    def get_serializer(self):
        return AttachmentsSerializer(
            data=self.request.data,
            context={
                "mode": self.mode,
                "user": self.user,
                "user_acl": self.user_acl,
                "post": self.post,
                "settings": self.settings,
            },
        )

    def save(self, serializer):
        serializer.save()


class AttachmentsSerializer(serializers.Serializer):
    attachments = serializers.ListField(
        child=serializers.IntegerField(), required=False
    )

    def __init__(self, *args, **kwargs):
        self.update_attachments = False
        self.removed_attachments = []
        self.final_attachments = []
        super().__init__(*args, **kwargs)

    def validate_attachments(self, ids):
        ids = list(set(ids))

        validate_attachments_count(ids, self.context["settings"])

        attachments = self.get_initial_attachments()
        new_attachments = self.get_new_attachments(ids)

        if not attachments and not new_attachments:
            return []  # no attachments

        # clean existing attachments
        for attachment in attachments:
            if attachment.pk in ids:
                self.final_attachments.append(attachment)
            else:
                if attachment.acl["can_delete"]:
                    self.update_attachments = True
                    self.removed_attachments.append(attachment)
                else:
                    message = pgettext(
                        "posting api",
                        'You don\'t have permission to remove "%(attachment)s" attachment.',
                    )
                    raise serializers.ValidationError(
                        message % {"attachment": attachment.filename}
                    )

        if new_attachments:
            self.update_attachments = True
            self.final_attachments += new_attachments
            self.final_attachments.sort(key=lambda a: a.pk, reverse=True)

    def get_initial_attachments(self):
        attachments = []
        if self.context["mode"] == PostingEndpoint.EDIT:
            queryset = self.context["post"].attachment_set.select_related("filetype")
            attachments = list(queryset)
            add_acl_to_obj(self.context["user_acl"], attachments)
        return attachments

    def get_new_attachments(self, ids):
        if not ids:
            return []

        queryset = (
            self.context["user"]
            .attachment_set.select_related("filetype")
            .filter(post__isnull=True, id__in=ids)
        )

        return list(queryset)

    def save(self):
        if not self.update_attachments:
            return

        if self.removed_attachments:
            for attachment in self.removed_attachments:
                attachment.delete_files()

            self.context["post"].attachment_set.filter(
                id__in=[a.id for a in self.removed_attachments]
            ).delete()

        if self.final_attachments:
            # sort final attachments by id, descending
            self.final_attachments.sort(key=lambda a: a.pk, reverse=True)
            self.context["user"].attachment_set.filter(
                id__in=[a.id for a in self.final_attachments]
            ).update(post=self.context["post"])

        self.sync_attachments_cache(self.context["post"], self.final_attachments)

    def sync_attachments_cache(self, post, attachments):
        if attachments:
            post.attachments_cache = AttachmentSerializer(attachments, many=True).data
            for attachment in post.attachments_cache:
                del attachment["acl"]
                del attachment["post"]
        else:
            post.attachments_cache = None
        post.update_fields.append("attachments_cache")


def validate_attachments_count(data, settings):
    total_attachments = len(data)
    if total_attachments > settings.post_attachments_limit:
        # pylint: disable=line-too-long
        message = npgettext(
            "posting api",
            "You can't attach more than %(limit_value)s file to single post (added %(show_value)s).",
            "You can't attach more than %(limit_value)s flies to single post (added %(show_value)s).",
            settings.post_attachments_limit,
        )
        raise serializers.ValidationError(
            message
            % {
                "limit_value": settings.post_attachments_limit,
                "show_value": total_attachments,
            }
        )



from datetime import timedelta

from django.contrib.auth import get_user_model
from django.urls import reverse
from django.utils import timezone

from ...acl.test import patch_user_acl
from ..models import Poll
from ..test import patch_category_acl
from .test_thread_poll_api import ThreadPollApiTestCase

User = get_user_model()


class ThreadGetVotesTests(ThreadPollApiTestCase):
    def setUp(self):
        super().setUp()

        self.mock_poll()

        self.poll.is_public = True
        self.poll.save()

        self.api_link = reverse(
            "misago:api:thread-poll-votes",
            kwargs={"thread_pk": self.thread.pk, "pk": self.poll.pk},
        )

    def test_anonymous(self):
        """api allows guests to get poll votes"""
        self.logout_user()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

    def test_invalid_thread_id(self):
        """api validates that thread id is integer"""
        api_link = reverse(
            "misago:api:thread-poll-votes",
            kwargs={"thread_pk": "kjha6dsa687sa", "pk": self.poll.pk},
        )

        response = self.client.get(api_link)
        self.assertEqual(response.status_code, 404)

    def test_nonexistant_thread_id(self):
        """api validates that thread exists"""
        api_link = reverse(
            "misago:api:thread-poll-votes",
            kwargs={"thread_pk": self.thread.pk + 1, "pk": self.poll.pk},
        )

        response = self.client.get(api_link)
        self.assertEqual(response.status_code, 404)

    def test_invalid_poll_id(self):
        """api validates that poll id is integer"""
        api_link = reverse(
            "misago:api:thread-poll-votes",
            kwargs={"thread_pk": self.thread.pk, "pk": "sad98as7d97sa98"},
        )

        response = self.client.get(api_link)
        self.assertEqual(response.status_code, 404)

    def test_nonexistant_poll_id(self):
        """api validates that poll exists"""
        api_link = reverse(
            "misago:api:thread-poll-votes",
            kwargs={"thread_pk": self.thread.pk, "pk": self.poll.pk + 123},
        )

        response = self.client.get(api_link)
        self.assertEqual(response.status_code, 404)

    @patch_user_acl({"can_always_see_poll_voters": False})
    def test_no_permission(self):
        """api chcecks permission to see poll voters"""
        self.poll.is_public = False
        self.poll.save()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)

    def test_nonpublic_poll(self):
        """api validates that poll is public"""
        self.logout_user()

        self.poll.is_public = False
        self.poll.save()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)

    def test_get_votes(self):
        """api returns list of voters"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(len(response_json), 4)

        self.assertEqual(
            [c["label"] for c in response_json], ["Alpha", "Beta", "Gamma", "Delta"]
        )
        self.assertEqual([c["votes"] for c in response_json], [1, 0, 2, 1])
        self.assertEqual([len(c["voters"]) for c in response_json], [1, 0, 2, 1])

        self.assertEqual(
            [[v["username"] for v in c["voters"]] for c in response_json][0][0], "User"
        )

        user = User.objects.get(slug="user")

        self.assertEqual(
            [[v["url"] for v in c["voters"]] for c in response_json][0][0],
            user.get_absolute_url(),
        )

    @patch_user_acl({"can_always_see_poll_voters": True})
    def test_get_votes_private_poll(self):
        """api returns list of voters on private poll for user with permission"""
        self.poll.is_public = False
        self.poll.save()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(len(response_json), 4)

        self.assertEqual(
            [c["label"] for c in response_json], ["Alpha", "Beta", "Gamma", "Delta"]
        )
        self.assertEqual([c["votes"] for c in response_json], [1, 0, 2, 1])
        self.assertEqual([len(c["voters"]) for c in response_json], [1, 0, 2, 1])

        self.assertEqual(
            [[v["username"] for v in c["voters"]] for c in response_json][0][0], "User"
        )

        user = User.objects.get(slug="user")

        self.assertEqual(
            [[v["url"] for v in c["voters"]] for c in response_json][0][0],
            user.get_absolute_url(),
        )


class ThreadPostVotesTests(ThreadPollApiTestCase):
    def setUp(self):
        super().setUp()

        self.mock_poll()

        self.api_link = reverse(
            "misago:api:thread-poll-votes",
            kwargs={"thread_pk": self.thread.pk, "pk": self.poll.pk},
        )

    def delete_user_votes(self):
        self.poll.choices[2]["votes"] = 1
        self.poll.choices[3]["votes"] = 0
        self.poll.votes = 2
        self.poll.save()

        self.poll.pollvote_set.filter(voter=self.user).delete()

    def test_anonymous(self):
        """api requires you to sign in to vote in poll"""
        self.logout_user()

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 403)

    def test_empty_vote_json(self):
        """api validates if vote that user has made was empty"""
        self.delete_user_votes()

        response = self.client.post(
            self.api_link, "[]", content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "You have to make a choice."})

    def test_empty_vote_form(self):
        """api validates if vote that user has made was empty"""
        self.delete_user_votes()

        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "You have to make a choice."})

    def test_malformed_vote(self):
        """api validates if vote that user has made was correctly structured"""
        self.delete_user_votes()

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": 'Expected a list of items but got type "dict".'}
        )

        response = self.post(self.api_link, data={})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": 'Expected a list of items but got type "dict".'}
        )

        response = self.post(self.api_link, data="hello")
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": 'Expected a list of items but got type "str".'}
        )

        response = self.post(self.api_link, data=123)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": 'Expected a list of items but got type "int".'}
        )

    def test_invalid_choices(self):
        """api validates if vote that user has made overlaps with allowed votes"""
        self.delete_user_votes()

        response = self.post(self.api_link, data=["lorem", "ipsum"])
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "One or more of poll choices were invalid."}
        )

    def test_too_many_choices(self):
        """api validates if vote that user has made overlaps with allowed votes"""
        self.poll.allowed_choices = 1
        self.poll.allow_revotes = True
        self.poll.save()

        response = self.post(self.api_link, data=["aaaaaaaaaaaa", "bbbbbbbbbbbb"])
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "This poll disallows voting for more than 1 choice."},
        )

    def test_revote(self):
        """
        api validates if user is trying to change vote in poll that disallows revoting
        """
        response = self.post(self.api_link, data=["lorem", "ipsum"])
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You have already voted in this poll."}
        )

        self.delete_user_votes()

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": 'Expected a list of items but got type "dict".'}
        )

    @patch_category_acl({"can_close_threads": False})
    def test_vote_in_closed_thread_no_permission(self):
        """api validates is user has permission to vote poll in closed thread"""
        self.thread.is_closed = True
        self.thread.save()

        self.delete_user_votes()

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This thread is closed. You can't vote in it."}
        )

    @patch_category_acl({"can_close_threads": True})
    def test_vote_in_closed_thread(self):
        """api validates is user has permission to vote poll in closed thread"""
        self.thread.is_closed = True
        self.thread.save()

        self.delete_user_votes()

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 400)

    @patch_category_acl({"can_close_threads": False})
    def test_vote_in_closed_category_no_permission(self):
        """api validates is user has permission to vote poll in closed category"""
        self.category.is_closed = True
        self.category.save()

        self.delete_user_votes()

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This category is closed. You can't vote in it."},
        )

    @patch_category_acl({"can_close_threads": True})
    def test_vote_in_closed_category(self):
        """api validates is user has permission to vote poll in closed category"""
        self.category.is_closed = True
        self.category.save()

        self.delete_user_votes()

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 400)

    def test_vote_in_finished_poll(self):
        """api valdiates if poll has finished before letting user to vote in it"""
        self.poll.posted_on = timezone.now() - timedelta(days=15)
        self.poll.length = 5
        self.poll.save()

        self.delete_user_votes()

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This poll is over. You can't vote in it."}
        )

        self.poll.length = 50
        self.poll.save()

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": 'Expected a list of items but got type "dict".'}
        )

    def test_fresh_vote(self):
        """api handles first vote in poll"""
        self.delete_user_votes()

        response = self.post(self.api_link, data=["aaaaaaaaaaaa", "bbbbbbbbbbbb"])
        self.assertEqual(response.status_code, 200)

        # validate state change
        poll = Poll.objects.get(pk=self.poll.pk)
        self.assertEqual(poll.votes, 4)
        self.assertEqual([c["votes"] for c in poll.choices], [2, 1, 1, 0])

        for choice in poll.choices:
            self.assertNotIn("selected", choice)

        self.assertEqual(poll.pollvote_set.count(), 4)

        # validate response json
        response_json = response.json()
        self.assertEqual(response_json["votes"], 4)
        self.assertEqual([c["votes"] for c in response_json["choices"]], [2, 1, 1, 0])
        self.assertEqual(
            [c["selected"] for c in response_json["choices"]],
            [True, True, False, False],
        )

        self.assertFalse(response_json["acl"]["can_vote"])

    def test_vote_change(self):
        """api handles vote change"""
        self.poll.allow_revotes = True
        self.poll.save()

        response = self.post(self.api_link, data=["aaaaaaaaaaaa", "bbbbbbbbbbbb"])
        self.assertEqual(response.status_code, 200)

        # validate state change
        poll = Poll.objects.get(pk=self.poll.pk)
        self.assertEqual(poll.votes, 4)
        self.assertEqual([c["votes"] for c in poll.choices], [2, 1, 1, 0])

        for choice in poll.choices:
            self.assertNotIn("selected", choice)

        self.assertEqual(poll.pollvote_set.count(), 4)

        # validate response json
        response_json = response.json()
        self.assertEqual(response_json["votes"], 4)
        self.assertEqual([c["votes"] for c in response_json["choices"]], [2, 1, 1, 0])
        self.assertEqual(
            [c["selected"] for c in response_json["choices"]],
            [True, True, False, False],
        )

        self.assertTrue(response_json["acl"]["can_vote"])



from .urls import *  # pylint: disable=wildcard-import, unused-wildcard-import

handler403 = "misago.core.testproject.views.mock_custom_403_error_page"
handler404 = "misago.core.testproject.views.mock_custom_404_error_page"



import os
import shutil

from django.core.files import File
from django.utils import timezone
from django.utils.crypto import get_random_string

from ...core.utils import slugify

FILENAME_MAX_LEN = 50


class DataArchive:
    def __init__(self, user, working_dir_path):
        self.user = user
        self.working_dir_path = working_dir_path

        self.tmp_dir_path = None
        self.data_dir_path = None

        self.file_path = None
        self.file = None

    def __enter__(self):
        self.tmp_dir_path = self.create_tmp_dir()
        self.data_dir_path = self.create_data_dir()

        return self

    def __exit__(self, *args):
        self.delete_file()
        self.delete_tmp_dir()

    def create_tmp_dir(self):
        tmp_dir_name = get_tmp_filename(self.user)
        tmp_dir_path = os.path.join(self.working_dir_path, tmp_dir_name)

        os.mkdir(tmp_dir_path)

        return tmp_dir_path

    def create_data_dir(self):
        data_dir_name = get_tmp_filename(self.user)
        data_dir_path = os.path.join(self.tmp_dir_path, data_dir_name)

        os.mkdir(data_dir_path)

        return data_dir_path

    def delete_tmp_dir(self):
        if self.tmp_dir_path:
            shutil.rmtree(self.tmp_dir_path)
            self.tmp_dir_path = None
            self.data_dir_path = None

    def get_file(self):
        file_name = get_tmp_filename(self.user)
        file_path = os.path.join(self.working_dir_path, file_name)

        self.file_path = shutil.make_archive(file_path, "zip", self.tmp_dir_path)
        self.file = open(self.file_path, "rb")

        return File(self.file)

    def delete_file(self):
        if self.file:
            self.file.close()
            self.file = None

        if self.file_path:
            os.remove(self.file_path)
            self.file_path = None

    def add_text(self, name, value, date=None, directory=None):
        clean_filename = slugify(str(name))
        file_dir_path = self.make_final_path(date=date, directory=directory)
        file_path = os.path.join(file_dir_path, "%s.txt" % clean_filename)
        with open(file_path, "w") as fp:
            fp.write(str(value))
            return file_path

    def add_dict(self, name, value, date=None, directory=None):
        text_lines = []
        for key, item in value.items():
            text_lines.append("%s: %s" % (key, item))
        text = "\n".join(text_lines)
        return self.add_text(name, text, date=date, directory=directory)

    def add_model_file(self, model_file, prefix=None, date=None, directory=None):
        if not model_file:
            return None

        target_dir_path = self.make_final_path(date=date, directory=directory)

        filename = os.path.basename(model_file.name)
        if prefix:
            prefixed_filename = "%s-%s" % (prefix, filename)
            clean_filename = trim_long_filename(prefixed_filename)
            target_path = os.path.join(target_dir_path, clean_filename)
        else:
            clean_filename = trim_long_filename(filename)
            target_path = os.path.join(target_dir_path, clean_filename)

        with open(target_path, "wb") as fp:
            for chunk in model_file.chunks():
                fp.write(chunk)

        return target_path

    def make_final_path(self, date=None, directory=None):
        # fixme: os.path.isdir test can be avoided in py37k
        if date and directory:
            raise ValueError("date and directory arguments are mutually exclusive")

        data_dir_path = self.data_dir_path

        if date:
            final_path = data_dir_path
            path_items = [date.strftime("%Y"), date.strftime("%m"), date.strftime("%d")]
            for path_item in path_items:
                final_path = os.path.join(final_path, str(path_item))
                if not os.path.isdir(final_path):
                    os.mkdir(final_path)
            return final_path

        if directory:
            final_path = os.path.join(data_dir_path, str(directory))
            if not os.path.isdir(final_path):
                os.mkdir(final_path)
            return final_path

        return data_dir_path


def get_tmp_filename(user):
    filename_bits = [
        user.slug,
        timezone.now().strftime("%Y%m%d-%H%M%S"),
        get_random_string(6),
    ]

    return "-".join(filename_bits)


def trim_long_filename(filename):
    # fixme: consider moving this utility to better place?
    # eg. to trim too long attachment filenames on upload
    if len(filename) < FILENAME_MAX_LEN:
        return filename

    name, extension = os.path.splitext(filename)
    name_len = FILENAME_MAX_LEN - len(extension)
    return "%s%s" % (name[:name_len], extension)



from pathlib import Path

import pytest
from django.core.files.base import ContentFile
from django.urls import reverse

from ....cache.test import assert_invalidates_cache
from ....test import assert_has_error_message
from ... import THEME_CACHE
from ...models import Theme, Css, Media


@pytest.fixture
def delete_link(theme):
    return reverse("misago:admin:themes:delete", kwargs={"pk": theme.pk})


def test_theme_without_children_can_be_deleted(admin_client, delete_link, theme):
    admin_client.post(delete_link)
    with pytest.raises(Theme.DoesNotExist):
        theme.refresh_from_db()


def test_theme_css_are_deleted_together_with_theme(admin_client, delete_link, css):
    admin_client.post(delete_link)
    with pytest.raises(Css.DoesNotExist):
        css.refresh_from_db()


def test_theme_source_css_files_are_deleted_together_with_theme(
    admin_client, delete_link, css
):
    admin_client.post(delete_link)
    assert not Path(css.source_file.path).exists()


def test_theme_build_css_files_are_deleted_together_with_theme(
    admin_client, delete_link, css
):
    css.build_file = ContentFile("body {}", name="test.css")
    css.build_hash = "abcdefgh"
    css.save()

    admin_client.post(delete_link)
    assert not Path(css.build_file.path).exists()


def test_theme_media_are_deleted_together_with_theme(admin_client, delete_link, media):
    admin_client.post(delete_link)
    with pytest.raises(Media.DoesNotExist):
        media.refresh_from_db()


def test_theme_images_are_deleted_together_with_theme(admin_client, delete_link, image):
    admin_client.post(delete_link)
    with pytest.raises(Media.DoesNotExist):
        image.refresh_from_db()


def test_theme_media_files_are_deleted_together_with_theme(
    admin_client, delete_link, media
):
    admin_client.post(delete_link)
    assert not Path(media.file.path).exists()


def test_theme_image_files_are_deleted_together_with_theme(
    admin_client, delete_link, image
):
    admin_client.post(delete_link)
    assert not Path(image.thumbnail.path).exists()


def test_theme_is_deleted_with_children(admin_client, delete_link, theme):
    Theme.objects.create(name="Child Theme", parent=theme)
    admin_client.post(delete_link)
    assert Theme.objects.count() == 1


def test_theme_children_are_deleted_recursively(admin_client, delete_link, theme):
    child_theme = Theme.objects.create(name="Child Theme", parent=theme)
    Theme.objects.create(name="Descendant Theme", parent=child_theme)
    Theme.objects.create(name="Descendant Theme", parent=child_theme)

    admin_client.post(delete_link)
    assert Theme.objects.count() == 1


def test_children_theme_can_be_deleted(admin_client, delete_link, theme, other_theme):
    theme.move_to(other_theme)
    theme.save()

    admin_client.post(delete_link)
    with pytest.raises(Theme.DoesNotExist):
        theme.refresh_from_db()


def test_deleting_children_theme_doesnt_delete_parent_themes(
    admin_client, delete_link, theme, other_theme
):
    theme.move_to(other_theme)
    theme.save()

    admin_client.post(delete_link)
    other_theme.refresh_from_db()


def test_deleting_theme_invalidates_themes_cache(admin_client, delete_link):
    with assert_invalidates_cache(THEME_CACHE):
        admin_client.post(delete_link)


def test_deleting_default_theme_sets_error_message(admin_client, default_theme):
    delete_link = reverse("misago:admin:themes:delete", kwargs={"pk": default_theme.pk})
    response = admin_client.post(delete_link)
    assert_has_error_message(response)


def test_default_theme_is_not_deleted(admin_client, default_theme):
    delete_link = reverse("misago:admin:themes:delete", kwargs={"pk": default_theme.pk})
    admin_client.post(delete_link)
    default_theme.refresh_from_db()


def test_deleting_active_theme_sets_error_message(admin_client, theme):
    theme.is_active = True
    theme.save()

    delete_link = reverse("misago:admin:themes:delete", kwargs={"pk": theme.pk})
    response = admin_client.post(delete_link)
    assert_has_error_message(response)


def test_active_theme_is_not_deleted(admin_client, theme):
    theme.is_active = True
    theme.save()

    delete_link = reverse("misago:admin:themes:delete", kwargs={"pk": theme.pk})
    admin_client.post(delete_link)
    theme.refresh_from_db()


def test_deleting_theme_containing_active_child_theme_sets_error_message(
    admin_client, theme, other_theme
):
    other_theme.move_to(theme)
    other_theme.is_active = True
    other_theme.save()

    delete_link = reverse("misago:admin:themes:delete", kwargs={"pk": theme.pk})
    response = admin_client.post(delete_link)
    assert_has_error_message(response)


def test_theme_containing_active_child_theme_is_not_deleted(
    admin_client, theme, other_theme
):
    other_theme.move_to(theme)
    other_theme.is_active = True
    other_theme.save()

    delete_link = reverse("misago:admin:themes:delete", kwargs={"pk": theme.pk})
    admin_client.post(delete_link)
    theme.refresh_from_db()



from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import Q
from django.utils import timezone
from django.utils.translation import pgettext_lazy

from ...conf import settings
from ...core.utils import slugify
from ...plugins.models import PluginDataModel


class Thread(PluginDataModel):
    WEIGHT_DEFAULT = 0
    WEIGHT_PINNED = 1
    WEIGHT_GLOBAL = 2

    WEIGHT_CHOICES = [
        (
            WEIGHT_DEFAULT,
            pgettext_lazy("thread weight choice", "Not pinned"),
        ),
        (
            WEIGHT_PINNED,
            pgettext_lazy("thread weight choice", "Pinned in category"),
        ),
        (
            WEIGHT_GLOBAL,
            pgettext_lazy("thread weight choice", "Pinned globally"),
        ),
    ]

    category = models.ForeignKey("misago_categories.Category", on_delete=models.CASCADE)
    title = models.CharField(max_length=255)
    slug = models.CharField(max_length=255)
    replies = models.PositiveIntegerField(default=0, db_index=True)

    has_events = models.BooleanField(default=False)
    has_poll = models.BooleanField(default=False)
    has_reported_posts = models.BooleanField(default=False)
    has_open_reports = models.BooleanField(default=False)
    has_unapproved_posts = models.BooleanField(default=False)
    has_hidden_posts = models.BooleanField(default=False)

    started_on = models.DateTimeField(db_index=True)
    last_post_on = models.DateTimeField(db_index=True)

    first_post = models.ForeignKey(
        "misago_threads.Post",
        related_name="+",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    starter = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=True, blank=True, on_delete=models.SET_NULL
    )
    starter_name = models.CharField(max_length=255)
    starter_slug = models.CharField(max_length=255)

    last_post = models.ForeignKey(
        "misago_threads.Post",
        related_name="+",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    last_post_is_event = models.BooleanField(default=False)
    last_poster = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="last_poster_set",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    last_poster_name = models.CharField(max_length=255, null=True, blank=True)
    last_poster_slug = models.CharField(max_length=255, null=True, blank=True)

    weight = models.PositiveIntegerField(default=WEIGHT_DEFAULT)

    is_unapproved = models.BooleanField(default=False, db_index=True)
    is_hidden = models.BooleanField(default=False)
    is_closed = models.BooleanField(default=False)

    best_answer = models.ForeignKey(
        "misago_threads.Post",
        related_name="+",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    best_answer_is_protected = models.BooleanField(default=False)
    best_answer_marked_on = models.DateTimeField(null=True, blank=True)
    best_answer_marked_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="marked_best_answer_set",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    best_answer_marked_by_name = models.CharField(max_length=255, null=True, blank=True)
    best_answer_marked_by_slug = models.CharField(max_length=255, null=True, blank=True)

    participants = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        related_name="privatethread_set",
        through="ThreadParticipant",
        through_fields=("thread", "user"),
    )

    class Meta:
        indexes = [
            *PluginDataModel.Meta.indexes,
            models.Index(
                name="misago_thread_pinned_glob_part",
                fields=["weight"],
                condition=Q(weight=2),
            ),
            models.Index(
                name="misago_thread_pinned_loca_part",
                fields=["weight"],
                condition=Q(weight=1),
            ),
            models.Index(
                name="misago_thread_not_pinned_part",
                fields=["weight"],
                condition=Q(weight=0),
            ),
            models.Index(
                name="misago_thread_not_global_part",
                fields=["weight"],
                condition=Q(weight__lt=2),
            ),
            models.Index(
                name="misago_thread_has_reporte_part",
                fields=["has_reported_posts"],
                condition=Q(has_reported_posts=True),
            ),
            models.Index(
                name="misago_thread_has_unappro_part",
                fields=["has_unapproved_posts"],
                condition=Q(has_unapproved_posts=True),
            ),
            models.Index(
                name="misago_thread_is_visible_part",
                fields=["is_hidden"],
                condition=Q(is_hidden=False),
            ),
            models.Index(fields=["category", "id"]),
            models.Index(fields=["category", "last_post_on"]),
            models.Index(fields=["category", "replies"]),
        ]

    def __str__(self):
        return self.title

    def delete(self, *args, **kwargs):
        from ..signals import delete_thread

        delete_thread.send(sender=self)

        super().delete(*args, **kwargs)

    def merge(self, other_thread):
        if self.pk == other_thread.pk:
            raise ValueError("thread can't be merged with itself")

        from ..signals import merge_thread

        merge_thread.send(sender=self, other_thread=other_thread)

    def move(self, new_category):
        from ..signals import move_thread

        self.category = new_category
        move_thread.send(sender=self)

    def synchronize(self):
        try:
            self.has_poll = bool(self.poll)
        except ObjectDoesNotExist:
            self.has_poll = False

        self.replies = self.post_set.filter(is_event=False, is_unapproved=False).count()

        if self.replies > 0:
            self.replies -= 1

        reported_post_qs = self.post_set.filter(has_reports=True)
        self.has_reported_posts = reported_post_qs.exists()

        if self.has_reported_posts:
            open_reports_qs = self.post_set.filter(has_open_reports=True)
            self.has_open_reports = open_reports_qs.exists()
        else:
            self.has_open_reports = False

        unapproved_post_qs = self.post_set.filter(is_unapproved=True)
        self.has_unapproved_posts = unapproved_post_qs.exists()

        hidden_post_qs = self.post_set.filter(is_hidden=True)[:1]
        self.has_hidden_posts = hidden_post_qs.exists()

        posts = self.post_set.order_by("id")

        first_post = posts.first()
        self.set_first_post(first_post)

        last_post = posts.filter(is_unapproved=False).last()
        if last_post:
            self.set_last_post(last_post)
        else:
            self.set_last_post(first_post)

        self.has_events = False
        if last_post:
            if last_post.is_event:
                self.has_events = True
            else:
                self.has_events = self.post_set.filter(is_event=True).exists()

    @property
    def has_best_answer(self):
        return bool(self.best_answer_id)

    @property
    def replies_in_ks(self):
        return "%sK" % round(self.replies / 1000, 0)

    @property
    def thread_type(self):
        return self.category.thread_type

    def get_api_url(self):
        return self.thread_type.get_thread_api_url(self)

    def get_editor_api_url(self):
        return self.thread_type.get_thread_editor_api_url(self)

    def get_merge_api_url(self):
        return self.thread_type.get_thread_merge_api_url(self)

    def get_posts_api_url(self):
        return self.thread_type.get_thread_posts_api_url(self)

    def get_post_merge_api_url(self):
        return self.thread_type.get_post_merge_api_url(self)

    def get_post_move_api_url(self):
        return self.thread_type.get_post_move_api_url(self)

    def get_post_split_api_url(self):
        return self.thread_type.get_post_split_api_url(self)

    def get_poll_api_url(self):
        return self.thread_type.get_thread_poll_api_url(self)

    def get_watch_api_url(self):
        return self.thread_type.get_thread_watch_api_url(self)

    def get_absolute_url(self, page=1):
        return self.thread_type.get_thread_absolute_url(self, page)

    def get_new_post_url(self):
        return self.thread_type.get_thread_new_post_url(self)

    def get_last_post_url(self):
        return self.thread_type.get_thread_last_post_url(self)

    def get_best_answer_url(self):
        return self.thread_type.get_thread_best_answer_url(self)

    def get_unapproved_post_url(self):
        return self.thread_type.get_thread_unapproved_post_url(self)

    def set_title(self, title):
        self.title = title
        self.slug = slugify(title)

        if self.id:
            from ..signals import update_thread_title

            update_thread_title.send(sender=self)

    def set_first_post(self, post):
        self.started_on = post.posted_on
        self.first_post = post
        self.starter = post.poster
        self.starter_name = post.poster_name
        if post.poster:
            self.starter_slug = post.poster.slug
        else:
            self.starter_slug = slugify(post.poster_name)

        self.is_unapproved = post.is_unapproved
        self.is_hidden = post.is_hidden

    def set_last_post(self, post):
        self.last_post_on = post.posted_on
        self.last_post_is_event = post.is_event
        self.last_post = post
        self.last_poster = post.poster
        self.last_poster_name = post.poster_name
        if post.poster:
            self.last_poster_slug = post.poster.slug
        else:
            self.last_poster_slug = slugify(post.poster_name)

    def set_best_answer(self, user, post):
        if post.thread_id != self.id:
            raise ValueError("post to set as best answer must be in same thread")
        if post.is_first_post:
            raise ValueError("post to set as best answer can't be first post")
        if post.is_hidden:
            raise ValueError("post to set as best answer can't be hidden")
        if post.is_unapproved:
            raise ValueError("post to set as best answer can't be unapproved")

        self.best_answer = post
        self.best_answer_is_protected = post.is_protected
        self.best_answer_marked_on = timezone.now()
        self.best_answer_marked_by = user
        self.best_answer_marked_by_name = user.username
        self.best_answer_marked_by_slug = user.slug

    def clear_best_answer(self):
        self.best_answer = None
        self.best_answer_is_protected = False
        self.best_answer_marked_on = None
        self.best_answer_marked_by = None
        self.best_answer_marked_by_name = None
        self.best_answer_marked_by_slug = None



# Generated by Django 1.11.1 on 2017-06-03 22:15
import django.contrib.postgres.fields
from django.contrib.postgres.operations import HStoreExtension
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("misago_users", "0009_redo_partial_indexes")]

    operations = [
        HStoreExtension(),
        migrations.AddField(
            model_name="user",
            name="profile_fields",
            field=django.contrib.postgres.fields.HStoreField(default=dict),
        ),
    ]



from ..outlets import PluginOutlet, append_outlet_action, template_outlet_action


@template_outlet_action
def noop_action(request, context):
    return None


def test_noop_template_action_renders_nothing(
    patch_outlets, render_outlet_template, snapshot
):
    append_outlet_action(PluginOutlet.TEST, noop_action)
    html = render_outlet_template()
    assert snapshot == html


@template_outlet_action
def template_name_action(request, context):
    return "misago/template_outlet_action_test.html"


def test_template_name_action_renders_template_with_standard_context(
    patch_outlets, render_outlet_template, snapshot
):
    append_outlet_action(PluginOutlet.TEST, template_name_action)
    html = render_outlet_template({"message": "Standard context"})
    assert snapshot == html


@template_outlet_action
def template_name_context_action(request, context):
    return "misago/template_outlet_action_test.html", {"message": "Custom context"}


def test_template_name_context_action_renders_template_with_updated_context(
    patch_outlets, render_outlet_template, snapshot
):
    append_outlet_action(PluginOutlet.TEST, template_name_context_action)
    html = render_outlet_template({"message": "Standard context"})
    assert snapshot == html



from django.urls import reverse

from ...conf.test import override_dynamic_settings
from ...core.utils import encode_json_html
from ..models import Ban
from ..test import UserTestCase, create_test_user
from ..tokens import make_password_change_token


class ForgottenPasswordViewsTests(UserTestCase):
    def test_guest_request_view_returns_200(self):
        """request new password view returns 200 for guests"""
        response = self.client.get(reverse("misago:forgotten-password"))
        self.assertEqual(response.status_code, 200)

    def test_authenticated_request_view_returns_200(self):
        """request new password view returns 200 for authenticated"""
        self.login_user(self.get_authenticated_user())

        response = self.client.get(reverse("misago:forgotten-password"))
        self.assertEqual(response.status_code, 200)

    def test_authenticated_request_unusable_password_view_returns_200(self):
        """
        request new password view returns 200 for authenticated with unusable password
        """
        user = self.get_authenticated_user()
        user.set_password(None)
        user.save()

        self.assertFalse(user.has_usable_password())
        self.login_user(user)

        response = self.client.get(reverse("misago:forgotten-password"))
        self.assertEqual(response.status_code, 200)

    def test_change_password_on_banned(self):
        """change banned user password errors"""
        user = create_test_user(
            "Other_User", "otheruser@example.com", self.USER_PASSWORD
        )

        Ban.objects.create(
            check_type=Ban.USERNAME, banned_value="Other_User", user_message="Nope!"
        )

        password_token = make_password_change_token(user)

        response = self.client.get(
            reverse(
                "misago:forgotten-password-change-form",
                kwargs={"pk": user.pk, "token": password_token},
            )
        )
        self.assertContains(response, encode_json_html("<p>Nope!</p>"), status_code=403)

    def test_change_password_on_other_user(self):
        """change other user password errors"""
        user = create_test_user(
            "Other_User", "otheruser@example.com", self.USER_PASSWORD
        )

        password_token = make_password_change_token(user)

        self.login_user(self.get_authenticated_user())

        response = self.client.get(
            reverse(
                "misago:forgotten-password-change-form",
                kwargs={"pk": user.pk, "token": password_token},
            )
        )
        self.assertContains(response, "your link has expired", status_code=400)

    def test_change_password_invalid_token(self):
        """invalid form token errors"""
        user = create_test_user(
            "Other_User", "otheruser@example.com", self.USER_PASSWORD
        )

        response = self.client.get(
            reverse(
                "misago:forgotten-password-change-form",
                kwargs={"pk": user.pk, "token": "abcdfghqsads"},
            )
        )
        self.assertContains(response, "your link is invalid", status_code=400)

    def test_change_password_form(self):
        """change user password form displays for valid token"""
        user = create_test_user(
            "Other_User", "otheruser@example.com", self.USER_PASSWORD
        )

        password_token = make_password_change_token(user)

        response = self.client.get(
            reverse(
                "misago:forgotten-password-change-form",
                kwargs={"pk": user.pk, "token": password_token},
            )
        )
        self.assertContains(response, password_token)

    def test_change_password_unusable_password_form(self):
        """set user first password form displays for valid token"""
        user = create_test_user("Other_User", "otheruser@example.com")

        password_token = make_password_change_token(user)

        response = self.client.get(
            reverse(
                "misago:forgotten-password-change-form",
                kwargs={"pk": user.pk, "token": password_token},
            )
        )
        self.assertContains(response, password_token)


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_provider="Lorem",
)
def test_forgotten_password_view_returns_403_if_oauth_is_enabled(user, client):
    password_token = make_password_change_token(user)

    response = client.get(
        reverse(
            "misago:forgotten-password-change-form",
            kwargs={"pk": user.pk, "token": password_token},
        )
    )
    assert response.status_code == 403



from unittest.mock import Mock

import pytest
from django.http import Http404

from ....notifications.models import Notification
from ..pagination import paginate_queryset


def test_pagination_with_before_returns_items_up_to_max_limit(notifications):
    request = Mock(GET={"before": notifications[10]})
    page = paginate_queryset(request, Notification.objects, "id", 4)

    assert len(page.items) == 4
    assert page.has_next
    assert page.has_previous

    items_verbs = [item.verb for item in page.items]
    assert items_verbs == ["test_6", "test_7", "test_8", "test_9"]


def test_pagination_with_before_returns_items_up_to_given_limit(notifications):
    request = Mock(GET={"before": notifications[10], "limit": 5})
    page = paginate_queryset(request, Notification.objects, "id", 10)

    assert len(page.items) == 5
    assert page.has_next
    assert page.has_previous

    items_verbs = [item.verb for item in page.items]
    assert items_verbs == ["test_5", "test_6", "test_7", "test_8", "test_9"]


def test_pagination_with_before_returns_last_items(notifications):
    request = Mock(GET={"before": notifications[5]})
    page = paginate_queryset(request, Notification.objects, "id", 10)

    assert len(page.items) == 5
    assert page.has_next
    assert not page.has_previous

    items_verbs = [item.verb for item in page.items]
    assert items_verbs == ["test_0", "test_1", "test_2", "test_3", "test_4"]


def test_pagination_with_before_raises_404_for_empty_item(notifications):
    request = Mock(GET={"before": notifications[0]})
    with pytest.raises(Http404):
        paginate_queryset(request, Notification.objects, "id", 10, raise_404=True)


def test_pagination_with_before_returns_items_up_to_max_limit_in_reverse_order(
    notifications,
):
    request = Mock(GET={"before": notifications[4]})
    page = paginate_queryset(request, Notification.objects, "-id", 4)

    assert len(page.items) == 4
    assert page.has_next
    assert page.has_previous

    items_verbs = [item.verb for item in page.items]
    assert items_verbs == ["test_8", "test_7", "test_6", "test_5"]


def test_pagination_with_before_returns_items_up_to_given_limit_in_reverse_order(
    notifications,
):
    request = Mock(GET={"before": notifications[4], "limit": 5})
    page = paginate_queryset(request, Notification.objects, "-id", 10)

    assert len(page.items) == 5
    assert page.has_next
    assert page.has_previous

    items_verbs = [item.verb for item in page.items]
    assert items_verbs == ["test_9", "test_8", "test_7", "test_6", "test_5"]


def test_pagination_with_before_returns_last_items_in_reverse_order(notifications):
    request = Mock(GET={"before": notifications[10]})
    page = paginate_queryset(request, Notification.objects, "-id", 10)

    assert len(page.items) == 4
    assert page.has_next
    assert not page.has_previous

    items_verbs = [item.verb for item in page.items]
    assert items_verbs == ["test_14", "test_13", "test_12", "test_11"]



from unittest.mock import Mock

from ..middleware import user_acl_middleware


def test_middleware_sets_attr_on_request(cache_versions, user):
    get_response = Mock()
    request = Mock(user=user, cache_versions=cache_versions)
    middleware = user_acl_middleware(get_response)
    middleware(request)
    assert request.user_acl


def test_middleware_calls_get_response(cache_versions, user):
    get_response = Mock()
    request = Mock(user=user, cache_versions=cache_versions)
    middleware = user_acl_middleware(get_response)
    middleware(request)
    get_response.assert_called_once()



from django.urls import reverse

from .. import test
from ...acl.test import patch_user_acl
from ..models import ThreadParticipant
from .test_privatethreads import PrivateThreadsTestCase


class PrivateThreadsListTests(PrivateThreadsTestCase):
    def setUp(self):
        super().setUp()

        self.test_link = reverse("misago:private-threads")

    def test_unauthenticated(self):
        """view requires user to sign in and be able to access it"""
        self.logout_user()

        response = self.client.get(self.test_link)
        self.assertContains(response, "sign in to use private threads", status_code=403)

    @patch_user_acl({"can_use_private_threads": False})
    def test_no_permission(self):
        """view requires user to have permission to be able to access it"""
        response = self.client.get(self.test_link)
        self.assertContains(response, "use private threads", status_code=403)

    def test_empty_list(self):
        """view has no showstoppers on returning empty list"""
        response = self.client.get(self.test_link)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "empty-message")

    def test_thread_visibility(self):
        """only participated threads are returned by private threads view"""
        visible = test.post_thread(category=self.category, poster=self.user)
        reported = test.post_thread(category=self.category, poster=self.user)

        # post hidden thread
        test.post_thread(category=self.category, poster=self.user)

        ThreadParticipant.objects.add_participants(visible, [self.user])

        reported.has_reported_posts = True
        reported.save()

        response = self.client.get(self.test_link)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, visible.get_absolute_url())

        # threads with reported posts will also show to moderators
        with patch_user_acl({"can_moderate_private_threads": True}):
            response = self.client.get(self.test_link)
            self.assertEqual(response.status_code, 200)
            self.assertContains(response, reported.get_absolute_url())
            self.assertContains(response, visible.get_absolute_url())



from typing import List

from .menuitems import get_footer_menu_items, get_navbar_menu_items
from .models import MenuItem


def menus(request):
    navbar_items = get_navbar_menu_items(request.cache_versions)
    footer_items = get_footer_menu_items(request.cache_versions)

    navbarItemsJson = serialize_items(navbar_items)
    footerItemsJson = []

    navbarUrls = [item["url"] for item in navbarItemsJson]
    for footerItem in serialize_items(footer_items):
        if footerItem["url"] not in navbarUrls:
            footerItemsJson.append(footerItem)

    request.frontend_context.update(
        {
            "extraMenuItems": navbarItemsJson,
            "extraFooterItems": footerItemsJson,
        }
    )

    return {
        "navbar_menu": navbar_items,
        "footer_menu": footer_items,
    }


def serialize_items(items: List[MenuItem]) -> List[dict]:
    serialized = []
    for item in items:
        serialized.append(
            {
                "title": item["title"],
                "url": item["url"],
                "className": item["css_class"],
                "targetBlank": item["target_blank"],
                "rel": item["rel"],
            }
        )
    return serialized



from io import StringIO

from django.core.management import call_command

from ...notifications.models import WatchedThread
from ..management.commands import createfakewatchedthreads


def test_management_command_creates_watched_threads(
    user, other_user, thread, other_thread
):
    call_command(createfakewatchedthreads.Command(), stdout=StringIO())
    assert WatchedThread.objects.count() == 4  # 2 users x 2 threads



import re

from django import forms
from django.core.validators import RegexValidator
from django.db import models
from django.utils.html import conditional_escape, mark_safe
from django.utils.translation import pgettext_lazy
from mptt.forms import TreeNodeChoiceField, TreeNodeMultipleChoiceField

from ...admin.forms import YesNoSwitch
from ...core.validators import validate_sluggable
from ...threads.threadtypes import trees_map
from .. import THREADS_ROOT_NAME
from ..models import Category, CategoryRole


class AdminCategoryFieldMixin:
    def __init__(self, *args, **kwargs):
        self.base_level = kwargs.pop("base_level", 1)
        kwargs["level_indicator"] = kwargs.get("level_indicator", "- - ")

        threads_tree_id = trees_map.get_tree_id_for_root(THREADS_ROOT_NAME)
        queryset = Category.objects.filter(tree_id=threads_tree_id)
        if not kwargs.pop("include_root", False):
            queryset = queryset.exclude(special_role="root_category")

        kwargs.setdefault("queryset", queryset)

        super().__init__(*args, **kwargs)

    def _get_level_indicator(self, obj):
        level = getattr(obj, obj._mptt_meta.level_attr) - self.base_level
        if level > 0:
            return mark_safe(conditional_escape(self.level_indicator) * level)
        return ""


class AdminCategoryChoiceField(AdminCategoryFieldMixin, TreeNodeChoiceField):
    pass


class AdminCategoryMultipleChoiceField(
    AdminCategoryFieldMixin, TreeNodeMultipleChoiceField
):
    pass


class CategoryFormBase(forms.ModelForm):
    name = forms.CharField(
        label=pgettext_lazy("admin category form", "Name"),
        validators=[validate_sluggable()],
    )
    short_name = forms.CharField(
        label=pgettext_lazy("admin category form", "Short name"),
        required=False,
        help_text=pgettext_lazy(
            "admin category form",
            "Optional, alternative or abbreviated name (eg. abbreviation) used on threads list.",
        ),
        validators=[validate_sluggable()],
    )
    color = forms.CharField(
        label=pgettext_lazy("admin category form", "Color"),
        required=False,
        help_text=pgettext_lazy(
            "admin category form",
            "Optional but recommended, should be in hex format, eg. #F5A9B8.",
        ),
        validators=[
            RegexValidator(
                r"^#[0-9a-f][0-9a-f][0-9a-f]([0-9a-f][0-9a-f][0-9a-f]?)$",
                flags=re.MULTILINE | re.IGNORECASE,
                message=pgettext_lazy(
                    "admin category form", "Entered value is not a valid color."
                ),
            ),
        ],
    )
    description = forms.CharField(
        label=pgettext_lazy("admin category form", "Description"),
        max_length=2048,
        required=False,
        widget=forms.Textarea(attrs={"rows": 3}),
        help_text=pgettext_lazy(
            "admin category form",
            "Optional description explaining category intented purpose.",
        ),
    )
    css_class = forms.CharField(
        label=pgettext_lazy("admin category form", "CSS class"),
        required=False,
        help_text=pgettext_lazy(
            "admin category form",
            "Optional CSS class used to customize this category's appearance from themes.",
        ),
    )
    is_closed = YesNoSwitch(
        label=pgettext_lazy("admin category form", "Closed category"),
        required=False,
        help_text=pgettext_lazy(
            "admin category form",
            "Only members with valid permissions can post in closed categories.",
        ),
    )
    require_threads_approval = YesNoSwitch(
        label=pgettext_lazy("admin category form", "Threads"),
        required=False,
        help_text=pgettext_lazy(
            "admin category form",
            "All threads started in this category will require moderator approval.",
        ),
    )
    require_replies_approval = YesNoSwitch(
        label=pgettext_lazy("admin category form", "Replies"),
        required=False,
        help_text=pgettext_lazy(
            "admin category form",
            "All replies posted in this category will require moderator approval.",
        ),
    )
    require_edits_approval = YesNoSwitch(
        label=pgettext_lazy("admin category form", "Edits"),
        required=False,
        help_text=pgettext_lazy(
            "admin category form",
            "Will make all edited replies return to unapproved state for moderator to review.",
        ),
    )
    prune_started_after = forms.IntegerField(
        label=pgettext_lazy("admin category form", "Thread age"),
        min_value=0,
        help_text=pgettext_lazy(
            "admin category form",
            "Prune thread if number of days since its creation is greater than specified. Enter 0 to disable this pruning criteria.",
        ),
    )
    prune_replied_after = forms.IntegerField(
        label=pgettext_lazy("admin category form", "Last reply"),
        min_value=0,
        help_text=pgettext_lazy(
            "admin category form",
            "Prune thread if number of days since last reply is greater than specified. Enter 0 to disable this pruning criteria.",
        ),
    )

    class Meta:
        model = Category
        fields = [
            "name",
            "short_name",
            "color",
            "description",
            "css_class",
            "is_closed",
            "require_threads_approval",
            "require_replies_approval",
            "require_edits_approval",
            "prune_started_after",
            "prune_replied_after",
            "archive_pruned_in",
        ]

    def clean_copy_permissions(self):
        data = self.cleaned_data["copy_permissions"]
        if data and data.pk == self.instance.pk:
            message = pgettext_lazy(
                "admin category form",
                "Permissions cannot be copied from category into itself.",
            )
            raise forms.ValidationError(message)
        return data

    def clean_archive_pruned_in(self):
        data = self.cleaned_data["archive_pruned_in"]
        if data and data.pk == self.instance.pk:
            message = pgettext_lazy(
                "admin category form", "Category cannot act as archive for itself."
            )
            raise forms.ValidationError(message)
        return data

    def clean(self):
        data = super().clean()
        self.instance.set_name(data.get("name"))
        return data


def CategoryFormFactory(instance):
    parent_queryset = Category.objects.all_categories(True).order_by("lft")
    if instance.pk:
        not_siblings = models.Q(lft__lt=instance.lft)
        not_siblings = not_siblings | models.Q(rght__gt=instance.rght)
        parent_queryset = parent_queryset.filter(not_siblings)

    return type(
        "CategoryFormFinal",
        (CategoryFormBase,),
        {
            "new_parent": AdminCategoryChoiceField(
                label=pgettext_lazy("admin category form", "Parent category"),
                queryset=parent_queryset,
                initial=instance.parent,
                empty_label=None,
            ),
            "copy_permissions": AdminCategoryChoiceField(
                label=pgettext_lazy("admin category form", "Copy permissions"),
                help_text=pgettext_lazy(
                    "admin category form",
                    "You can replace this category permissions with permissions copied from category selected here.",
                ),
                queryset=Category.objects.all_categories(),
                empty_label=pgettext_lazy(
                    "admin category form", "Don't copy permissions"
                ),
                required=False,
            ),
            "archive_pruned_in": AdminCategoryChoiceField(
                label=pgettext_lazy("admin category form", "Archive"),
                help_text=pgettext_lazy(
                    "admin category form",
                    "Instead of being deleted, pruned threads can be moved to designated category.",
                ),
                queryset=Category.objects.all_categories(),
                empty_label=pgettext_lazy(
                    "admin category form", "Don't archive pruned threads"
                ),
                required=False,
            ),
        },
    )


class DeleteCategoryFormBase(forms.ModelForm):
    class Meta:
        model = Category
        fields = []

    def clean(self):
        data = super().clean()

        if data.get("move_threads_to"):
            if data["move_threads_to"].pk == self.instance.pk:
                message = pgettext_lazy(
                    "admin category form",
                    "You are trying to move this category threads to itself.",
                )
                raise forms.ValidationError(message)

            moving_to_child = self.instance.has_child(data["move_threads_to"])
            if moving_to_child and not data.get("move_children_to"):
                message = pgettext_lazy(
                    "admin category form",
                    "You are trying to move this category threads to a child category that will also be deleted.",
                )
                raise forms.ValidationError(message)

        return data


def DeleteFormFactory(instance):
    content_queryset = Category.objects.all_categories().order_by("lft")
    fields = {
        "move_threads_to": AdminCategoryChoiceField(
            label=pgettext_lazy("admin category form", "Move category threads to"),
            queryset=content_queryset,
            initial=instance.parent,
            empty_label=pgettext_lazy("admin category form", "Delete with category"),
            required=False,
        )
    }

    not_siblings = models.Q(lft__lt=instance.lft)
    not_siblings = not_siblings | models.Q(rght__gt=instance.rght)
    children_queryset = Category.objects.all_categories(True)
    children_queryset = children_queryset.filter(not_siblings).order_by("lft")

    if children_queryset.exists():
        fields["move_children_to"] = AdminCategoryChoiceField(
            label=pgettext_lazy("admin category form", "Move child categories to"),
            queryset=children_queryset,
            empty_label=pgettext_lazy("admin category form", "Delete with category"),
            required=False,
        )

    return type("DeleteCategoryFormFinal", (DeleteCategoryFormBase,), fields)


class CategoryRoleForm(forms.ModelForm):
    name = forms.CharField(label=pgettext_lazy("admin category role form", "Role name"))

    class Meta:
        model = CategoryRole
        fields = ["name"]


def RoleCategoryACLFormFactory(category, category_roles, selected_role):
    attrs = {
        "category": category,
        "role": forms.ModelChoiceField(
            label=pgettext_lazy("admin permissions form", "Role"),
            required=False,
            queryset=category_roles,
            initial=selected_role,
            empty_label=pgettext_lazy("admin permissions form", "No access"),
        ),
    }

    return type("RoleCategoryACLForm", (forms.Form,), attrs)


def CategoryRolesACLFormFactory(role, category_roles, selected_role):
    attrs = {
        "role": role,
        "category_role": forms.ModelChoiceField(
            label=pgettext_lazy("admin permissions form", "Role"),
            required=False,
            queryset=category_roles,
            initial=selected_role,
            empty_label=pgettext_lazy("admin permissions form", "No access"),
        ),
    }

    return type("CategoryRolesACLForm", (forms.Form,), attrs)



from ..useracl import get_user_acl


def test_getter_returns_authenticated_user_acl(cache_versions, user):
    acl = get_user_acl(user, cache_versions)

    assert acl
    assert acl["user_id"] == user.id
    assert acl["is_authenticated"] is True
    assert acl["is_anonymous"] is False


def test_user_acl_includes_staff_and_superuser_false_status(cache_versions, user):
    acl = get_user_acl(user, cache_versions)

    assert acl
    assert acl["is_staff"] is False
    assert acl["is_superuser"] is False


def test_user_acl_includes_cache_versions(cache_versions, user):
    acl = get_user_acl(user, cache_versions)

    assert acl
    assert acl["cache_versions"] == cache_versions


def test_getter_returns_anonymous_user_acl(db, cache_versions, anonymous_user):
    acl = get_user_acl(anonymous_user, cache_versions)

    assert acl
    assert acl["user_id"] == anonymous_user.id
    assert acl["is_authenticated"] is False
    assert acl["is_anonymous"] is True


def test_superuser_acl_includes_staff_and_superuser_true_status(
    cache_versions, superuser
):
    acl = get_user_acl(superuser, cache_versions)

    assert acl
    assert acl["is_staff"] is True
    assert acl["is_superuser"] is True


def test_staffuser_acl_includes_staff_and_superuser_true_status(
    cache_versions, staffuser
):
    acl = get_user_acl(staffuser, cache_versions)

    assert acl
    assert acl["is_staff"] is True
    assert acl["is_superuser"] is False


def test_getter_returns_acl_from_cache(mocker, db, cache_versions, anonymous_user):
    cache_get = mocker.patch("django.core.cache.cache.get", return_value=dict())
    get_user_acl(anonymous_user, cache_versions)
    cache_get.assert_called_once()


def test_getter_builds_new_acl_when_cache_is_not_available(
    mocker, cache_versions, user
):
    mocker.patch("django.core.cache.cache.set")
    mocker.patch("misago.acl.buildacl.build_acl", return_value=dict())
    cache_get = mocker.patch("django.core.cache.cache.get", return_value=None)

    get_user_acl(user, cache_versions)
    cache_get.assert_called_once()


def test_getter_sets_new_cache_if_no_cache_is_set(
    mocker, db, cache_versions, anonymous_user
):
    cache_set = mocker.patch("django.core.cache.cache.set")
    mocker.patch("misago.acl.buildacl.build_acl", return_value=dict())
    mocker.patch("django.core.cache.cache.get", return_value=None)

    get_user_acl(anonymous_user, cache_versions)
    cache_set.assert_called_once()


def test_acl_cache_name_includes_cache_version(
    mocker, db, cache_versions, anonymous_user
):
    cache_set = mocker.patch("django.core.cache.cache.set")
    mocker.patch("misago.acl.buildacl.build_acl", return_value=dict())
    mocker.patch("django.core.cache.cache.get", return_value=None)

    get_user_acl(anonymous_user, cache_versions)
    cache_key = cache_set.call_args[0][0]
    assert cache_versions["acl"] in cache_key


def test_getter_is_not_setting_new_cache_if_cache_is_set(
    mocker, cache_versions, anonymous_user
):
    cache_set = mocker.patch("django.core.cache.cache.set")
    mocker.patch("django.core.cache.cache.get", return_value=dict())

    get_user_acl(anonymous_user, cache_versions)
    cache_set.assert_not_called()



from ..categories import (
    fake_category,
    fake_category_description,
    fake_category_name,
    fake_closed_category,
)


def test_fake_category_can_be_created(fake, root_category):
    assert fake_category(fake, root_category)


def test_fake_category_is_created_with_specified_parent(fake, default_category):
    category = fake_category(fake, default_category)
    assert category.parent == default_category


def test_fake_category_can_be_created_with_copy_of_other_category_acls(
    fake, root_category, default_category
):
    category = fake_category(fake, root_category, copy_acl_from=default_category)
    for acl in default_category.category_role_set.all():
        category.category_role_set.get(role=acl.role, category_role=acl.category_role)


def test_fake_closed_category_can_be_created(fake, root_category):
    category = fake_closed_category(fake, root_category)
    assert category.is_closed


def test_fake_category_name_can_be_created(fake):
    assert fake_category_name(fake)


def test_different_fake_category_name_is_created_every_time(fake):
    fake_names = [fake_category_name(fake) for i in range(5)]
    assert len(fake_names) == len(set(fake_names))


def test_fake_category_description_can_be_created(fake):
    assert fake_category_description(fake)


def test_different_fake_category_description_is_created_every_time(fake):
    fake_descriptions = [fake_category_description(fake) for i in range(5)]
    assert len(fake_descriptions) == len(set(fake_descriptions))



from django.db import models
from django.utils import timezone

from ...conf import settings


class Subscription(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    thread = models.ForeignKey("Thread", on_delete=models.CASCADE)
    category = models.ForeignKey("misago_categories.Category", on_delete=models.CASCADE)

    last_read_on = models.DateTimeField(default=timezone.now)
    send_email = models.BooleanField(default=False)

    class Meta:
        indexes = [
            models.Index(fields=["send_email", "last_read_on"]),
        ]
